"""Lark parser for Jac Lang."""

from __future__ import annotations

import ast as py_ast
import os
from typing import Optional, Sequence, TypeAlias, TypeVar

# from icecream import ic

import jaclang.compiler.absyntree as ast
from jaclang.compiler.constant import Tokens as Tok
from jaclang.compiler.passes.ir_pass import Pass
from jaclang.utils.helpers import pascal_to_snake

T = TypeVar("T", bound=ast.AstNode)


class PyastBuildPass(Pass[ast.PythonModuleAst]):
    """Jac Parser."""

    def __init__(self, input_ir: ast.PythonModuleAst) -> None:
        """Initialize parser."""
        self.mod_path = input_ir.loc.mod_path
        Pass.__init__(self, input_ir=input_ir, prior=None)

    def nu(self, node: T) -> T:
        """Update node."""
        self.cur_node = node
        return node

    def pp(self, node: py_ast.AST) -> None:
        """Print python node."""
        # print(
        #     f"{node.__class__.__name__} - {[(k, type(v)) for k, v in vars(node).items()]}"
        # )

    def convert(self, node: py_ast.AST) -> ast.AstNode:
        """Get python node type."""
        # print(
        #     f"working on {type(node).__name__} line {node.lineno if hasattr(node, 'lineno') else 0}"
        # )
        if hasattr(self, f"proc_{pascal_to_snake(type(node).__name__)}"):
            ret = getattr(self, f"proc_{pascal_to_snake(type(node).__name__)}")(node)
        else:
            raise self.ice(f"Unknown node type {type(node).__name__}")
        # print(f"finshed {type(node).__name__} ---------------------")
        # print("normalizing", ret.__class__.__name__)
        # ic("normalizing", ret.__class__.__name__)
        # print(ret.unparse())
        # ret.unparse()
        return ret

    def transform(self, ir: ast.PythonModuleAst) -> ast.Module:
        """Transform input IR."""
        self.ir: ast.Module = self.proc_module(ir.ast)
        return self.ir

    def extract_with_entry(
        self, body: list[ast.AstNode], exclude_types: TypeAlias = T
    ) -> list[T | ast.ModuleCode]:
        """Extract with entry from a body."""

        def gen_mod_code(with_entry_body: list[ast.CodeBlockStmt]) -> ast.ModuleCode:
            with_entry_subnodelist = ast.SubNodeList[ast.CodeBlockStmt](
                items=with_entry_body,
                delim=Tok.WS,
                kid=with_entry_body,
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
            return ast.ModuleCode(
                name=None,
                body=with_entry_subnodelist,
                kid=[with_entry_subnodelist],
                doc=None,
            )

        extracted: list[T | ast.ModuleCode] = []
        with_entry_body: list[ast.CodeBlockStmt] = []
        for i in body:
            if isinstance(i, exclude_types):
                if len(with_entry_body):
                    extracted.append(gen_mod_code(with_entry_body))
                    with_entry_body = []
                extracted.append(i)
            elif isinstance(i, ast.CodeBlockStmt):
                if isinstance(i, ast.ExprStmt) and isinstance(i.expr, ast.String):
                    self.convert_to_doc(i.expr)
                with_entry_body.append(i)
            else:
                continue  # FIXME: check this
                # self.ice("Invalid type for with entry body")
        if len(with_entry_body):
            extracted.append(gen_mod_code(with_entry_body))
        return extracted

    def proc_module(self, node: py_ast.Module) -> ast.Module:
        """Process python node.

        class Module(mod):
            __match_args__ = ("body", "type_ignores")
            body: list[stmt]
            type_ignores: list[TypeIgnore]
        """
        elements: list[ast.AstNode] = [self.convert(i) for i in node.body]
        elements[0] = (
            elements[0].expr
            if isinstance(elements[0], ast.ExprStmt)
            and isinstance(elements[0].expr, ast.String)
            else elements[0]
        )
        doc_str_list = [elements[0]] if isinstance(elements[0], ast.String) else []
        valid = (
            (doc_str_list)
            + self.extract_with_entry(elements[1:], (ast.ElementStmt, ast.EmptyToken))
            if doc_str_list
            else self.extract_with_entry(elements[:], (ast.ElementStmt, ast.EmptyToken))
        )
        doc_str = elements[0] if isinstance(elements[0], ast.String) else None
        self.convert_to_doc(doc_str) if doc_str else None
        ret = ast.Module(
            name=self.mod_path.split(os.path.sep)[-1].split(".")[0],
            source=ast.JacSource("", mod_path=self.mod_path),
            doc=doc_str,
            body=valid[1:] if valid and isinstance(valid[0], ast.String) else valid,
            is_imported=False,
            kid=valid,
        )
        return self.nu(ret)

    def proc_function_def(
        self, node: py_ast.FunctionDef | py_ast.AsyncFunctionDef
    ) -> ast.Ability:
        """Process python node.

        class FunctionDef(stmt):
            name: _Identifier
            args: arguments
            body: list[stmt]
            decorator_list: list[expr]
            returns: expr | None
            if sys.version_info >= (3, 12):
            type_params: list[type_param]
        """
        name = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=(
                node.name if node.name != "root" else "root_"
            ),  # root is a reserved keyword
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.name),
            pos_start=0,
            pos_end=0,
        )
        body = [self.convert(i) for i in node.body]
        valid = [i for i in body if isinstance(i, (ast.CodeBlockStmt))]
        if len(valid) != len(body):
            raise self.ice("Length mismatch in function body")

        if (
            len(valid)
            and isinstance(valid[0], ast.ExprStmt)
            and isinstance(valid[0].expr, ast.String)
        ):
            self.convert_to_doc(valid[0].expr)
            doc = valid[0]
            valid_body = ast.SubNodeList[ast.CodeBlockStmt](
                items=[doc] + valid[1:],
                delim=Tok.WS,
                kid=valid[1:] + [doc],
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
        else:
            doc = None
            valid_body = ast.SubNodeList[ast.CodeBlockStmt](
                items=valid,
                delim=Tok.WS,
                kid=valid,
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
        decorators = [self.convert(i) for i in node.decorator_list]
        valid_dec = [i for i in decorators if isinstance(i, ast.Expr)]
        if len(valid_dec) != len(decorators):
            raise self.ice("Length mismatch in decorators on function")
        valid_decorators = (
            ast.SubNodeList[ast.Expr](
                items=valid_dec, delim=Tok.DECOR_OP, kid=decorators
            )
            if len(valid_dec)
            else None
        )
        res = self.convert(node.args)
        sig: Optional[ast.FuncSignature] = (
            res if isinstance(res, ast.FuncSignature) else None
        )
        ret_sig = self.convert(node.returns) if node.returns else None
        if isinstance(ret_sig, ast.Expr):
            if not sig:
                sig = ast.FuncSignature(params=None, return_type=ret_sig, kid=[ret_sig])
            else:
                sig.return_type = ret_sig
                sig.add_kids_right([sig.return_type])
        kid = ([doc] if doc else []) + (
            [name, sig, valid_body] if sig else [name, valid_body]
        )
        if not sig:
            raise self.ice("Function signature not found")
        ret = ast.Ability(
            name_ref=name,
            is_async=False,
            is_static=False,
            is_abstract=False,
            is_override=False,
            access=None,
            signature=sig,
            body=valid_body,
            decorators=valid_decorators,
            doc=None,
            kid=kid,
        )
        return ret

    def proc_async_function_def(self, node: py_ast.AsyncFunctionDef) -> ast.Ability:
        """Process python node.

        class AsyncFunctionDef(stmt):
            __match_args__ = ("name", "args", "body", "decorator_list",
                              "returns", "type_comment", "type_params")
            name: _Identifier
            args: arguments
            body: list[stmt]
            decorator_list: list[expr]
            returns: expr | None
            if sys.version_info >= (3, 12):
                type_params: list[type_param]
        """
        ability = self.proc_function_def(node)
        ability.is_async = True
        return ability

    def proc_class_def(self, node: py_ast.ClassDef) -> ast.Architype | ast.Enum:
        """Process python node.

        class ClassDef(stmt):
            __match_args__ = ("name", "bases", "keywords", "body",
                              "decorator_list", "type_params")
            name: _Identifier
            bases: list[expr]
            keywords: list[keyword]
            body: list[stmt]
            decorator_list: list[expr]
            if sys.version_info >= (3, 12):
            type_params: list[type_param]
        """
        name = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.name,
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.name),
            pos_start=0,
            pos_end=0,
        )
        arch_type = ast.Token(
            file_path=self.mod_path,
            name=Tok.KW_CLASS,
            value="class",
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=0,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )
        body = [self.convert(i) for i in node.body]
        for body_stmt in body:
            if (
                isinstance(body_stmt, ast.Ability)
                and isinstance(body_stmt.name_ref, ast.Name)
                and body_stmt.name_ref.value == "__init__"
            ):
                tok = ast.Name(
                    file_path=self.mod_path,
                    name=Tok.KW_INIT,
                    value="init",
                    line=node.lineno,
                    end_line=node.end_lineno if node.end_lineno else node.lineno,
                    col_start=node.col_offset,
                    col_end=node.col_offset + len("init"),
                    pos_start=0,
                    pos_end=0,
                )
                body_stmt.name_ref = ast.SpecialVarRef(var=tok)
            if (
                isinstance(body_stmt, ast.Ability)
                and body_stmt.signature
                and isinstance(body_stmt.signature, ast.FuncSignature)
                and body_stmt.signature.params
            ):
                body_stmt.signature.params.items = [
                    param
                    for param in body_stmt.signature.params.items
                    if param.name.value != "self"
                ]
        doc = (
            body[0].expr
            if isinstance(body[0], ast.ExprStmt)
            and isinstance(body[0].expr, ast.String)
            else None
        )
        self.convert_to_doc(doc) if doc else None
        body = body[1:] if doc else body
        valid: list[ast.ArchBlockStmt] = (
            self.extract_with_entry(body, ast.ArchBlockStmt)
            if body and not (isinstance(body[0], ast.Semi) and len(body) == 1)
            else []
        )
        empty_block: Sequence[ast.AstNode] = [
            self.operator(Tok.LBRACE, "{"),
            self.operator(Tok.RBRACE, "}"),
        ]
        valid_body = ast.SubNodeList[ast.ArchBlockStmt](
            items=valid,
            delim=Tok.WS,
            kid=(valid if valid else empty_block),
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        converted_base_classes = [self.convert(base) for base in node.bases]
        base_classes: list[ast.Expr] = [
            base for base in converted_base_classes if isinstance(base, ast.Expr)
        ]
        valid_bases = (
            ast.SubNodeList[ast.Expr](
                items=base_classes, delim=Tok.COMMA, kid=base_classes
            )
            if base_classes
            else None
        )
        converted_decorators_list = [self.convert(i) for i in node.decorator_list]
        decorators = [i for i in converted_decorators_list if isinstance(i, ast.Expr)]
        valid_decorators = (
            ast.SubNodeList[ast.Expr](
                items=decorators, delim=Tok.DECOR_OP, kid=decorators
            )
            if decorators
            else None
        )
        if (
            base_classes
            and isinstance(base_classes[0], ast.Name)
            and base_classes[0].value == "Enum"
        ):
            if len(base_classes) > 1:
                raise ValueError(
                    "Python's Enum class cannot be used with multiple inheritance."
                )
            arch_type.name = Tok.KW_ENUM
            arch_type.value = "enum"
            valid_enum_body: list[ast.EnumBlockStmt] = []
            for class_body_stmt in node.body:
                converted_stmt = self.convert(class_body_stmt)
                if isinstance(converted_stmt, ast.EnumBlockStmt):
                    if isinstance(converted_stmt, ast.Assignment):
                        converted_stmt.is_enum_stmt = True
                    valid_enum_body.append(converted_stmt)
                else:
                    if isinstance(converted_stmt, ast.ExprStmt) and isinstance(
                        converted_stmt.expr, ast.String
                    ):
                        continue
                    pintok = ast.Token(
                        file_path=self.mod_path,
                        name=Tok.PYNLINE,
                        value=py_ast.unparse(class_body_stmt),
                        line=node.lineno,
                        end_line=node.end_lineno if node.end_lineno else node.lineno,
                        col_start=node.col_offset,
                        col_end=node.col_offset + len(py_ast.unparse(class_body_stmt)),
                        pos_start=0,
                        pos_end=0,
                    )
                    valid_enum_body.append(ast.PyInlineCode(code=pintok, kid=[pintok]))

            valid_enum_body2: list[ast.EnumBlockStmt] = [
                i for i in valid_enum_body if isinstance(i, ast.EnumBlockStmt)
            ]
            enum_body = (
                ast.SubNodeList[ast.EnumBlockStmt](
                    items=valid_enum_body2, delim=Tok.COMMA, kid=valid_enum_body2
                )
                if valid_enum_body2
                else None
            )
            if doc:
                doc.line_no = name.line_no
            return ast.Enum(
                name=name,
                access=None,
                base_classes=None,
                body=enum_body,
                kid=(
                    [doc, name, enum_body]
                    if doc and enum_body
                    else (
                        [doc, name]
                        if doc
                        else [name, enum_body] if enum_body else [name]
                    )
                ),
                doc=doc,
                decorators=valid_decorators,
            )
        kid = (
            [name, valid_bases, valid_body, doc]
            if doc and valid_bases
            else (
                [name, valid_bases, valid_body]
                if valid_bases
                else [name, valid_body, doc] if doc else [name, valid_body]
            )
        )
        return ast.Architype(
            arch_type=arch_type,
            name=name,
            access=None,
            base_classes=valid_bases,
            body=valid_body,
            kid=kid,
            doc=doc,
            decorators=valid_decorators,
        )

    def proc_return(self, node: py_ast.Return) -> ast.ReturnStmt | None:
        """Process python node.

        class Return(stmt):
            __match_args__ = ("value",)
            value: expr | None
        """
        value = self.convert(node.value) if node.value else None
        if not value:
            return ast.ReturnStmt(
                expr=None, kid=[self.operator(Tok.KW_RETURN, "return")]
            )
        elif value and isinstance(value, ast.Expr):
            return ast.ReturnStmt(expr=value, kid=[value])
        else:
            raise self.ice("Invalid return value")

    def proc_delete(self, node: py_ast.Delete) -> ast.DeleteStmt:
        """Process python node.

        class Delete(stmt):
            __match_args__ = ("targets",)
            targets: list[expr]
        """
        exprs = [self.convert(target) for target in node.targets]
        valid_exprs = [expr for expr in exprs if isinstance(expr, ast.Expr)]
        if not len(valid_exprs) or len(valid_exprs) != len(exprs):
            raise self.ice("Length mismatch in delete targets")
        target = ast.SubNodeList[ast.Expr | ast.KWPair](
            items=[*valid_exprs], delim=Tok.COMMA, kid=exprs
        )
        target_1 = (
            valid_exprs[0]
            if len(valid_exprs) > 1
            else ast.TupleVal(values=target, kid=[target])
        )
        return ast.DeleteStmt(
            target=target_1,
            kid=[target],
        )

    def proc_assign(self, node: py_ast.Assign) -> ast.Assignment:
        """Process python node.

        class Assign(stmt):
            targets: list[expr]
            value: expr
        """
        value = self.convert(node.value)
        targets = [self.convert(target) for target in node.targets]
        valid = [target for target in targets if isinstance(target, ast.Expr)]
        if len(valid) == len(targets):
            valid_targets = ast.SubNodeList[ast.Expr](
                items=valid, delim=Tok.EQ, kid=valid
            )
        else:
            raise self.ice("Length mismatch in assignment targets")
        if isinstance(value, ast.Expr):
            return ast.Assignment(
                target=valid_targets,
                value=value,
                type_tag=None,
                kid=[valid_targets, value],
            )
        else:
            raise self.ice()

    def proc_aug_assign(self, node: py_ast.AugAssign) -> ast.Assignment:
        """Process python node.

        class AugAssign(stmt):
            __match_args__ = ("target", "op", "value")
            target: Name | Attribute | Subscript
            op: operator
            value: expr
        """
        from jaclang.compiler import TOKEN_MAP

        target = self.convert(node.target)
        op = self.convert(node.op)
        if isinstance(op, ast.Token):
            op.name = self.aug_op_map(TOKEN_MAP, op)
        value = self.convert(node.value)
        if (
            isinstance(value, ast.Expr)
            and isinstance(target, ast.Expr)
            and isinstance(op, ast.Token)
        ):
            targ = ast.SubNodeList[ast.Expr](
                items=[target], delim=Tok.COMMA, kid=[target]
            )
            return ast.Assignment(
                target=targ,
                type_tag=None,
                mutable=True,
                aug_op=op,
                value=value,
                kid=[target, op, value],
            )
        else:
            raise self.ice()

    def proc_ann_assign(self, node: py_ast.AnnAssign) -> ast.Assignment:
        """Process python node.

        class AnnAssign(stmt):
            __match_args__ = ("target", "annotation", "value", "simple")
            target: Name | Attribute | Subscript
            annotation: expr
            value: expr | None
            simple: int
        """
        target = self.convert(node.target)
        annotation = self.convert(node.annotation)
        if isinstance(annotation, ast.Expr):
            annotation_subtag = ast.SubTag[ast.Expr](tag=annotation, kid=[annotation])
        else:
            raise self.ice()
        value = self.convert(node.value) if node.value else None
        if (
            (isinstance(value, (ast.Expr, ast.YieldExpr)) or not value)
            and isinstance(annotation, ast.Expr)
            and isinstance(target, ast.Expr)
        ):
            target = ast.SubNodeList[ast.Expr](
                items=[target], delim=Tok.EQ, kid=[target]
            )
            return ast.Assignment(
                target=target,
                value=value if isinstance(value, (ast.Expr, ast.YieldExpr)) else None,
                type_tag=annotation_subtag,
                kid=(
                    [target, annotation_subtag, value]
                    if value
                    else [target, annotation_subtag]
                ),
            )
        else:
            raise self.ice()

    def proc_for(self, node: py_ast.For) -> ast.InForStmt:
        """Process python node.

        class For(stmt):
            __match_args__ = ("target", "iter", "body", "orelse")
            target: expr
            iter: expr
            body: list[stmt]
            orelse: list[stmt]
        """
        target = self.convert(node.target)
        iter = self.convert(node.iter)
        body = [self.convert(i) for i in node.body]
        val_body = [i for i in body if isinstance(i, ast.CodeBlockStmt)]
        if len(val_body) != len(body):
            raise self.ice("Length mismatch in for body")

        valid_body = ast.SubNodeList[ast.CodeBlockStmt](
            items=val_body,
            delim=Tok.WS,
            kid=val_body,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        orelse = [self.convert(i) for i in node.orelse]
        val_orelse = [i for i in orelse if isinstance(i, ast.CodeBlockStmt)]
        if len(val_orelse) != len(orelse):
            raise self.ice("Length mismatch in for orelse")
        if orelse:
            valid_orelse = ast.SubNodeList[ast.CodeBlockStmt](
                items=val_orelse,
                delim=Tok.WS,
                kid=orelse,
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
        else:
            valid_orelse = None
        fin_orelse = (
            ast.ElseStmt(body=valid_orelse, kid=[valid_orelse])
            if valid_orelse
            else None
        )
        if isinstance(target, ast.Expr) and isinstance(iter, ast.Expr):
            return ast.InForStmt(
                target=target,
                is_async=False,
                collection=iter,
                body=valid_body,
                else_body=fin_orelse,
                kid=(
                    [target, iter, valid_body, fin_orelse]
                    if fin_orelse
                    else [target, iter, valid_body]
                ),
            )
        else:
            raise self.ice()

    def proc_async_for(self, node: py_ast.AsyncFor) -> ast.InForStmt:
        """Process AsyncFor node.

        class AsyncFor(stmt):
            __match_args__ = ("target", "iter", "body", "orelse")
            target: expr
            iter: expr
            body: list[stmt]
            orelse: list[stmt]`
        """
        target = self.convert(node.target)
        iter = self.convert(node.iter)
        body = [self.convert(i) for i in node.body]
        val_body = [i for i in body if isinstance(i, ast.CodeBlockStmt)]
        if len(val_body) != len(body):
            raise self.ice("Length mismatch in for body")

        valid_body = ast.SubNodeList[ast.CodeBlockStmt](
            items=val_body,
            delim=Tok.WS,
            kid=val_body,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        orelse = [self.convert(i) for i in node.orelse]
        val_orelse = [i for i in orelse if isinstance(i, ast.CodeBlockStmt)]
        if len(val_orelse) != len(orelse):
            raise self.ice("Length mismatch in for orelse")
        if orelse:
            valid_orelse = ast.SubNodeList[ast.CodeBlockStmt](
                items=val_orelse,
                delim=Tok.WS,
                kid=orelse,
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
        else:
            valid_orelse = None
        fin_orelse = (
            ast.ElseStmt(body=valid_orelse, kid=[valid_orelse])
            if valid_orelse
            else None
        )
        if isinstance(target, ast.Expr) and isinstance(iter, ast.Expr):
            return ast.InForStmt(
                target=target,
                is_async=True,
                collection=iter,
                body=valid_body,
                else_body=fin_orelse,
                kid=(
                    [target, iter, valid_body, fin_orelse]
                    if fin_orelse
                    else [target, iter, valid_body]
                ),
            )
        else:
            raise self.ice()

    def proc_while(self, node: py_ast.While) -> ast.WhileStmt:
        """Process While node.

        class While(stmt):
            __match_args__ = ("test", "body", "orelse")
            test: expr
            body: list[stmt]
            orelse: list[stmt]
        """
        test = self.convert(node.test)
        body = [self.convert(stmt) for stmt in node.body]
        valid_body = [stmt for stmt in body if isinstance(stmt, ast.CodeBlockStmt)]
        if len(valid_body) != len(body):
            raise self.ice("Length mismatch in while body")
        fin_body = ast.SubNodeList[ast.CodeBlockStmt](
            items=valid_body,
            delim=Tok.WS,
            kid=valid_body,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        if isinstance(test, ast.Expr):
            return ast.WhileStmt(
                condition=test,
                body=fin_body,
                kid=[test, fin_body],
            )
        else:
            raise self.ice()

    def proc_if(self, node: py_ast.If) -> ast.IfStmt:
        """Process If node.

        class If(stmt):
            __match_args__ = ("test", "body", "orelse")
            test: expr
            body: list[stmt]
            orelse: list[stmt]
        """
        test = self.convert(node.test)
        body = [self.convert(stmt) for stmt in node.body]
        valid_body = [stmt for stmt in body if isinstance(stmt, ast.CodeBlockStmt)]
        if len(valid_body) != len(body):
            self.error("Length mismatch in async for body")
        body2 = ast.SubNodeList[ast.CodeBlockStmt](
            items=valid_body,
            delim=Tok.WS,
            kid=body,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )

        orelse = [self.convert(stmt) for stmt in node.orelse]
        valid_orelse = [
            stmt for stmt in orelse if isinstance(stmt, (ast.CodeBlockStmt))
        ]
        if valid_orelse:
            first_elm = valid_orelse[0]
            if isinstance(first_elm, ast.IfStmt):
                else_body: Optional[ast.ElseIf | ast.ElseStmt] = ast.ElseIf(
                    condition=first_elm.condition,
                    body=first_elm.body,
                    else_body=first_elm.else_body,
                    kid=first_elm.kid,
                )
            else:
                orelse2 = ast.SubNodeList[ast.CodeBlockStmt](
                    items=valid_orelse,
                    delim=Tok.WS,
                    kid=orelse,
                    left_enc=self.operator(Tok.LBRACE, "{"),
                    right_enc=self.operator(Tok.RBRACE, "}"),
                )
                else_body = ast.ElseStmt(body=orelse2, kid=[orelse2])
        else:
            else_body = None
        if isinstance(test, ast.Expr):
            ret = ast.IfStmt(
                condition=test,
                body=body2,
                else_body=else_body,
                kid=([test, body2, else_body] if else_body else [test, body2]),
            )
        else:
            raise self.ice()
        return ret

    def proc_with(self, node: py_ast.With) -> ast.WithStmt:
        """Process With node.

        class With(stmt):
            __match_args__ = ("items", "body")
            items: list[withitem]
            body: list[stmt]
        """
        items = [self.convert(item) for item in node.items]
        valid_items = [item for item in items if isinstance(item, ast.ExprAsItem)]
        if len(valid_items) != len(items):
            raise self.ice("Length mismatch in with items")
        items_sub = ast.SubNodeList[ast.ExprAsItem](
            items=valid_items, delim=Tok.COMMA, kid=items
        )
        body = [self.convert(stmt) for stmt in node.body]
        valid_body = [stmt for stmt in body if isinstance(stmt, ast.CodeBlockStmt)]
        if len(valid_body) != len(body):
            raise self.ice("Length mismatch in async for body")
        body_sub = ast.SubNodeList[ast.CodeBlockStmt](
            items=valid_body,
            delim=Tok.WS,
            kid=body,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        return ast.WithStmt(
            is_async=False, exprs=items_sub, body=body_sub, kid=[items_sub, body_sub]
        )

    def proc_async_with(self, node: py_ast.AsyncWith) -> ast.WithStmt:
        """Process AsyncWith node.

        class AsyncWith(stmt):
            __match_args__ = ("items", "body")
            items: list[withitem]
            body: list[stmt]
        """
        items = [self.convert(item) for item in node.items]
        valid_items = [item for item in items if isinstance(item, ast.ExprAsItem)]
        if len(valid_items) != len(items):
            raise self.ice("Length mismatch in with items")
        items_sub = ast.SubNodeList[ast.ExprAsItem](
            items=valid_items, delim=Tok.COMMA, kid=items
        )
        body = [self.convert(stmt) for stmt in node.body]
        valid_body = [stmt for stmt in body if isinstance(stmt, ast.CodeBlockStmt)]
        if len(valid_body) != len(body):
            raise self.ice("Length mismatch in async for body")
        body_sub = ast.SubNodeList[ast.CodeBlockStmt](
            items=valid_body,
            delim=Tok.WS,
            kid=body,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        return ast.WithStmt(
            is_async=True, exprs=items_sub, body=body_sub, kid=[items_sub, body_sub]
        )

    def proc_raise(self, node: py_ast.Raise) -> ast.RaiseStmt:
        """Process python node.

        class Raise(stmt):
            exc: expr | None
            cause: expr | None
        """
        exc = self.convert(node.exc) if node.exc else None
        cause = self.convert(node.cause) if node.cause else None
        kid: list[ast.Expr | ast.Token] = []
        if isinstance(exc, ast.Expr):
            kid = [exc]
        if isinstance(cause, ast.Expr):
            kid.append(cause)
        if not (exc and cause):
            kid.append(self.operator(Tok.KW_RAISE, "raise"))
        if (isinstance(cause, ast.Expr) or cause is None) and (
            isinstance(exc, ast.Expr) or exc is None
        ):
            if exc and not node.cause:
                return ast.RaiseStmt(
                    cause=exc,
                    from_target=None,
                    kid=[self.operator(Tok.KW_RAISE, "raise"), exc],
                )
            else:
                return ast.RaiseStmt(cause=cause, from_target=exc, kid=kid)
        else:
            raise self.ice()

    def proc_assert(self, node: py_ast.Assert) -> ast.AssertStmt:
        """Process python node.

        class Assert(stmt):
            test: expr
            msg: expr | None
        """
        test = self.convert(node.test)
        msg = self.convert(node.msg) if node.msg else None
        if isinstance(test, ast.Expr) and (isinstance(msg, ast.Expr) or msg is None):
            return ast.AssertStmt(
                condition=test,
                error_msg=msg,
                kid=[test, msg] if msg else [test],
            )
        else:
            raise self.ice()

    def proc_attribute(self, node: py_ast.Attribute) -> ast.AtomTrailer:
        """Proassignment targetscess python node.

        class Attribute(expr):
        if sys.version_info >= (3, 10):
            __match_args__ = ("value", "attr", "ctx")
        value: expr
        attr: _Identifier
        ctx: expr_context
        """
        value = self.convert(node.value)
        if (
            isinstance(value, ast.FuncCall)
            and isinstance(value.target, ast.Name)
            and value.target.value == "super"
        ):
            tok = ast.Name(
                file_path=self.mod_path,
                name=Tok.KW_SUPER,
                value="super",
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + len("super"),
                pos_start=0,
                pos_end=0,
            )
            value = ast.SpecialVarRef(var=tok)
            # exit()
        attribute = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=(
                ("<>" + node.attr)
                if node.attr == "init"
                else "init" if node.attr == "__init__" else node.attr
            ),
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.attr),
            pos_start=0,
            pos_end=0,
        )
        if isinstance(value, ast.Expr):
            return ast.AtomTrailer(
                target=value,
                right=attribute,
                is_attr=True,
                is_null_ok=False,
                kid=[value, attribute],
            )
        else:
            raise self.ice()

    def proc_await(self, node: py_ast.Await) -> ast.AwaitExpr:
        """Process python node.

        class Await(expr):
            value: expr
        """
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            return ast.AwaitExpr(target=value, kid=[value])
        else:
            raise self.ice()

    def proc_bin_op(self, node: py_ast.BinOp) -> ast.AtomUnit:
        """Process python node.

        class BinOp(expr):
            if sys.version_info >= (3, 10):
                __match_args__ = ("left", "op", "right")
            left: expr
            op: operator
            right: expr
        """
        left = self.convert(node.left)
        op = self.convert(node.op)
        right = self.convert(node.right)
        if (
            isinstance(left, ast.Expr)
            and isinstance(op, ast.Token)
            and isinstance(right, ast.Expr)
        ):
            value = ast.BinaryExpr(
                left=left,
                op=op,
                right=right,
                kid=[left, op, right],
            )
            return ast.AtomUnit(
                value=value,
                kid=[
                    self.operator(Tok.RPAREN, "("),
                    value,
                    self.operator(Tok.LPAREN, ")"),
                ],
            )
        else:
            raise self.ice()

    def proc_unary_op(self, node: py_ast.UnaryOp) -> ast.UnaryExpr:
        """Process python node.

        class UnaryOp(expr):
            op: unaryop
            operand: expr
        """
        op = self.convert(node.op)
        operand = self.convert(node.operand)
        if isinstance(op, ast.Token) and isinstance(operand, ast.Expr):
            return ast.UnaryExpr(
                op=op,
                operand=operand,
                kid=[op, operand],
            )
        else:
            raise self.ice()

    def proc_bool_op(self, node: py_ast.BoolOp) -> ast.BoolExpr:
        """Process python node.

        class BoolOp(expr): a and b and c
            op: boolop
            values: list[expr]
        """
        op = self.convert(node.op)
        values = [self.convert(value) for value in node.values]
        valid = [value for value in values if isinstance(value, ast.Expr)]
        valid_values = ast.SubNodeList[ast.Expr](
            items=valid, delim=Tok.COMMA, kid=values
        )
        if isinstance(op, ast.Token) and len(valid) == len(values):
            return ast.BoolExpr(op=op, values=valid, kid=[op, valid_values])
        else:
            raise self.ice()

    def proc_break(self, node: py_ast.Break) -> ast.CtrlStmt:
        """Process python node."""
        break_tok = ast.Token(
            file_path=self.mod_path,
            name=Tok.KW_BREAK,
            value="break",
            line=0,
            end_line=0,
            col_start=0,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )
        return ast.CtrlStmt(ctrl=break_tok, kid=[break_tok])

    def proc_call(self, node: py_ast.Call) -> ast.FuncCall:
        """Process python node.

        class Call(expr):
        if sys.version_info >= (3, 10):
            __match_args__ = ("func", "args", "keywords")
        func: expr
        args: list[expr]
        keywords: list[keyword]
        """
        func = self.convert(node.func)
        params_in: list[ast.Expr | ast.KWPair] = []
        args = [self.convert(arg) for arg in node.args]
        keywords = [self.convert(keyword) for keyword in node.keywords]

        for i in args:
            if isinstance(i, ast.Expr):
                params_in.append(i)
        for i in keywords:
            if isinstance(i, ast.KWPair):
                params_in.append(i)
        if len(params_in) != 0:
            params_in2 = ast.SubNodeList[ast.Expr | ast.KWPair](
                items=params_in, delim=Tok.COMMA, kid=params_in
            )
        else:
            params_in2 = None
        if isinstance(func, ast.Expr):
            return ast.FuncCall(
                target=func,
                params=params_in2,
                genai_call=None,
                kid=[func, params_in2] if params_in2 else [func],
            )
        else:
            raise self.ice()

    def proc_compare(self, node: py_ast.Compare) -> ast.CompareExpr:
        """Process python node.

        class Compare(expr):
        if sys.version_info >= (3, 10):
            __match_args__ = ("left", "ops", "comparators")
        left: expr
        ops: list[cmpop]
        comparators: list[expr]
        """
        left = self.convert(node.left)
        comparators = [self.convert(comparator) for comparator in node.comparators]
        valid_comparators = [
            comparator for comparator in comparators if isinstance(comparator, ast.Expr)
        ]
        comparators2 = ast.SubNodeList[ast.Expr](
            items=valid_comparators, delim=Tok.COMMA, kid=comparators
        )
        ops = [self.convert(op) for op in node.ops]
        valid_ops = [op for op in ops if isinstance(op, ast.Token)]
        ops2 = ast.SubNodeList[ast.Token](items=valid_ops, delim=Tok.COMMA, kid=ops)

        kids = [left, ops2, comparators2]
        if (
            isinstance(left, ast.Expr)
            and len(ops) == len(valid_ops)
            and len(comparators) == len(valid_comparators)
        ):
            return ast.CompareExpr(
                left=left, rights=valid_comparators, ops=valid_ops, kid=kids
            )
        else:
            raise self.ice()

    def proc_constant(self, node: py_ast.Constant) -> ast.Literal:
        """Process python node.

        class Constant(expr):
            value: Any  # None, str, bytes, bool, int, float, complex, Ellipsis
            kind: str | None
            # Aliases for value, for backwards compatibility
            s: Any
            n: int | float | complex
        """
        type_mapping = {
            int: ast.Int,
            float: ast.Float,
            str: ast.String,
            bytes: ast.String,
            bool: ast.Bool,
            type(None): ast.Null,
        }
        value_type = type(node.value)
        if value_type in type_mapping:
            if value_type is None:
                token_type = "NULL"
            elif value_type == str:
                token_type = "STRING"
            else:
                token_type = f"{value_type.__name__.upper()}"

            return type_mapping[value_type](
                file_path=self.mod_path,
                name=token_type,
                value=(
                    f'"{repr(node.value)[1:-1]}"'
                    if value_type == str
                    else str(node.value)
                ),
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + len(str(node.value)),
                pos_start=0,
                pos_end=0,
            )
        elif node.value == Ellipsis:
            return ast.Ellipsis(
                file_path=self.mod_path,
                name=Tok.ELLIPSIS,
                value="...",
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + 3,
                pos_start=0,
                pos_end=0,
            )
        else:
            raise self.ice("Invalid type for constant")

    def proc_continue(self, node: py_ast.Continue) -> ast.CtrlStmt:
        """Process python node."""
        continue_tok = ast.Token(
            file_path=self.mod_path,
            name=Tok.KW_CONTINUE,
            value="continue",
            line=0,
            end_line=0,
            col_start=0,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )
        return ast.CtrlStmt(ctrl=continue_tok, kid=[continue_tok])

    def proc_dict(self, node: py_ast.Dict) -> ast.DictVal:
        """Process python node.

        class Dict(expr):
            keys: list[expr | None]
            values: list[expr]
        """
        keys = [self.convert(i) if i else None for i in node.keys]
        values = [self.convert(i) for i in node.values]
        valid_keys = [i for i in keys if isinstance(i, ast.Expr) or i is None]
        valid_values = [i for i in values if isinstance(i, ast.Expr)]
        kvpair: list[ast.KVPair] = []
        for i in range(len(values)):
            key_pluck = valid_keys[i]
            kvp = ast.KVPair(
                key=key_pluck,
                value=valid_values[i],
                kid=([key_pluck, valid_values[i]] if key_pluck else [valid_values[i]]),
            )
            kvpair.append(kvp)
        return ast.DictVal(
            kv_pairs=kvpair,
            kid=(
                [*kvpair]
                if len(kvpair)
                else [self.operator(Tok.LBRACE, "{"), self.operator(Tok.RBRACE, "}")]
            ),
        )

    def proc_dict_comp(self, node: py_ast.DictComp) -> ast.DictCompr:
        """Process python node.

        class DictComp(expr):
            key: expr
            value: expr
            generators: list[comprehension]
        """
        key = self.convert(node.key)
        value = self.convert(node.value)
        if isinstance(key, ast.Expr) and isinstance(value, ast.Expr):
            kv_pair = ast.KVPair(key=key, value=value, kid=[key, value])
        else:
            raise self.ice()
        generators = [self.convert(i) for i in node.generators]
        valid = [i for i in generators if isinstance(i, (ast.InnerCompr))]
        if len(valid) != len(generators):
            raise self.ice("Length mismatch in dict compr generators")
        compr = ast.SubNodeList[ast.InnerCompr](items=valid, delim=Tok.COMMA, kid=valid)
        return ast.DictCompr(kv_pair=kv_pair, compr=valid, kid=[kv_pair, compr])

    def proc_ellipsis(self, node: py_ast.Ellipsis) -> None:
        """Process python node."""

    def proc_except_handler(self, node: py_ast.ExceptHandler) -> ast.Except:
        """Process python node.

        class ExceptHandler(excepthandler):
            type: expr | None
            name: _Identifier | None
            body: list[stmt]
        """
        type = self.convert(node.type) if node.type else None
        name: ast.Name | None = None
        if not type and not node.name:
            type = ast.Name(
                file_path=self.mod_path,
                name=Tok.NAME,
                value="Exception",
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + 9,
                pos_start=0,
                pos_end=0,
            )
            name = ast.Name(
                file_path=self.mod_path,
                name=Tok.NAME,
                value="e",
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + 1,
                pos_start=0,
                pos_end=0,
            )
        else:
            # type = ast.Name(
            #     file_path=self.mod_path,
            #     name=Tok.NAME,
            #     value=no,
            #     line=node.lineno,
            #     end_line = (node.end_lineno if node.end_lineno else node.lineno,)
            #     col_start=node.col_offset,
            #     col_end=node.col_offset + 9,
            #     pos_start=0,
            #     pos_end=0,
            # )
            name = (
                ast.Name(
                    file_path=self.mod_path,
                    name=Tok.NAME,
                    value=node.name,
                    line=node.lineno,
                    end_line=node.end_lineno if node.end_lineno else node.lineno,
                    col_start=node.col_offset,
                    col_end=node.col_offset + len(node.name),
                    pos_start=0,
                    pos_end=0,
                )
                if node.name
                else None
            )

        body = [self.convert(i) for i in node.body]
        valid = [i for i in body if isinstance(i, (ast.CodeBlockStmt))]
        if len(valid) != len(body):
            raise self.ice("Length mismatch in except handler body")
        valid_body = ast.SubNodeList[ast.CodeBlockStmt](
            items=valid,
            delim=Tok.WS,
            kid=valid,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        kid = [item for item in [type, name, valid_body] if item]
        if isinstance(type, ast.Expr) and (isinstance(name, ast.Name) or not name):
            return ast.Except(
                ex_type=type,
                name=name,
                body=valid_body,
                kid=kid,
            )
        else:
            raise self.ice()

    def proc_expr(self, node: py_ast.Expr) -> ast.ExprStmt:
        """Process python node.

        class Expr(stmt):
            value: expr
        """
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            return ast.ExprStmt(expr=value, in_fstring=False, kid=[value])
        else:
            raise self.ice()

    def proc_formatted_value(self, node: py_ast.FormattedValue) -> ast.ExprStmt:
        """Process python node.

        class FormattedValue(expr):
        if sys.version_info >= (3, 10):
            __match_args__ = ("value", "conversion", "format_spec")
        value: expr
        conversion: int
        format_spec: expr | None
        """
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            ret = ast.ExprStmt(
                expr=value,
                in_fstring=True,
                kid=[value],
            )
        else:
            raise self.ice()
        return ret

    def proc_function_type(self, node: py_ast.FunctionType) -> None:
        """Process python node."""

    def proc_generator_exp(self, node: py_ast.GeneratorExp) -> ast.GenCompr:
        """Process python node..

        class SetComp(expr):
            elt: expr
            generators: list[comprehension]
        """
        elt = self.convert(node.elt)
        generators = [self.convert(gen) for gen in node.generators]
        valid = [gen for gen in generators if isinstance(gen, ast.InnerCompr)]
        if len(generators) != len(valid):
            raise self.ice("Length mismatch in list comp generators")
        compr = ast.SubNodeList[ast.InnerCompr](items=valid, delim=Tok.COMMA, kid=valid)
        if isinstance(elt, ast.Expr):
            return ast.GenCompr(out_expr=elt, compr=valid, kid=[elt, compr])
        else:
            raise self.ice()

    def proc_global(self, node: py_ast.Global) -> ast.GlobalStmt:
        """Process python node.

        class Global(stmt):
            names: list[_Identifier]
        """
        names: list[ast.NameAtom] = []
        for id in node.names:
            names.append(
                ast.Name(
                    file_path=self.mod_path,
                    name=Tok.NAME,
                    value=id,
                    line=node.lineno,
                    end_line=node.end_lineno if node.end_lineno else node.lineno,
                    col_start=node.col_offset,
                    col_end=node.col_offset + len(id),
                    pos_start=0,
                    pos_end=0,
                )
            )
        target = ast.SubNodeList[ast.NameAtom](items=names, delim=Tok.COMMA, kid=names)
        return ast.GlobalStmt(target=target, kid=[target])

    def proc_if_exp(self, node: py_ast.IfExp) -> ast.IfElseExpr:
        """Process python node.

        class IfExp(expr):
            test: expr
            body: expr
            orelse: expr
        """
        test = self.convert(node.test)
        body = self.convert(node.body)
        orelse = self.convert(node.orelse)
        if (
            isinstance(test, ast.Expr)
            and isinstance(body, ast.Expr)
            and isinstance(orelse, ast.Expr)
        ):
            return ast.IfElseExpr(
                value=body, condition=test, else_value=orelse, kid=[body, test, orelse]
            )
        else:
            raise self.ice()

    def proc_import(self, node: py_ast.Import) -> ast.Import:
        """Process python node.

        class Import(stmt):
            names: list[alias]
        """
        names = [self.convert(name) for name in node.names]
        valid_names = [name for name in names if isinstance(name, ast.ExprAsItem)]
        if len(valid_names) != len(names):
            self.error("Length mismatch in import names")
        paths = []
        for name in valid_names:
            if isinstance(name.expr, ast.Name) and (
                isinstance(name.alias, ast.Name) or name.alias is None
            ):
                paths.append(
                    ast.ModulePath(
                        path=[name.expr],
                        level=0,
                        alias=name.alias,
                        kid=[i for i in name.kid if i],
                    )
                )
            # Need to unravel atom trailers
            else:
                raise self.ice()
        lang = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value="py",
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )
        pytag = ast.SubTag[ast.Name](tag=lang, kid=[lang])
        items = ast.SubNodeList[ast.ModulePath](items=paths, delim=Tok.COMMA, kid=paths)
        ret = ast.Import(
            hint=pytag,
            from_loc=None,
            items=items,
            is_absorb=False,
            kid=[pytag, items],
        )
        return ret

    def proc_import_from(self, node: py_ast.ImportFrom) -> ast.Import:
        """Process python node.

        class ImportFrom(stmt):
            module: str | None
            names: list[alias]
            level: int
        """
        lang = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value="py",
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )
        modpaths: list[ast.Name] = []
        if node.module:
            for i in node.module.split("."):
                modpaths.append(
                    ast.Name(
                        file_path=self.mod_path,
                        name=Tok.NAME,
                        value=i,
                        line=node.lineno,
                        end_line=node.end_lineno if node.end_lineno else node.lineno,
                        col_start=0,
                        col_end=0,
                        pos_start=0,
                        pos_end=0,
                    )
                )
        path = ast.ModulePath(
            path=modpaths,
            level=node.level,
            alias=None,
            kid=modpaths,
        )
        names = [self.convert(name) for name in node.names]
        valid_names = []
        for name in names:
            if (
                isinstance(name, ast.ExprAsItem)
                and isinstance(name.expr, ast.Name)
                and (isinstance(name.alias, ast.Name) or name.alias is None)
            ):
                valid_names.append(
                    ast.ModuleItem(
                        name=name.expr,
                        alias=name.alias if name.alias else None,
                        kid=[i for i in name.kid if i],
                    )
                )
            else:
                raise self.ice()
        items = (
            ast.SubNodeList[ast.ModuleItem](
                items=valid_names, delim=Tok.COMMA, kid=valid_names
            )
            if valid_names
            else None
        )
        if not items:
            raise self.ice("No valid names in import from")
        pytag = ast.SubTag[ast.Name](tag=lang, kid=[lang])
        if len(node.names) == 1 and node.names[0].name == "*":
            path_in = ast.SubNodeList[ast.ModulePath](
                items=[path], delim=Tok.COMMA, kid=[path]
            )
            ret = ast.Import(
                hint=pytag,
                from_loc=None,
                items=path_in,
                is_absorb=True,
                kid=[pytag, path_in],
            )
            return ret
        ret = ast.Import(
            hint=pytag,
            from_loc=path,
            items=items,
            is_absorb=False,
            kid=[pytag, path, items],
        )
        return ret

    def proc_joined_str(self, node: py_ast.JoinedStr) -> ast.FString:
        """Process python node.

        class JoinedStr(expr):
        if sys.version_info >= (3, 10):
            __match_args__ = ("values",)
        values: list[expr]
        """
        values = [self.convert(value) for value in node.values]
        valid = [
            value for value in values if isinstance(value, (ast.String, ast.ExprStmt))
        ]
        valid_values = ast.SubNodeList[ast.String | ast.ExprStmt](
            items=valid, delim=None, kid=valid
        )
        return ast.FString(parts=valid_values, kid=[valid_values])

    def proc_lambda(self, node: py_ast.Lambda) -> ast.LambdaExpr:
        """Process python node.

        class Lambda(expr):
            args: arguments
            body: expr
        """
        args = self.convert(node.args)
        body = self.convert(node.body)
        if isinstance(args, ast.FuncSignature) and isinstance(body, ast.Expr):
            return ast.LambdaExpr(signature=args, body=body, kid=[args, body])
        else:
            raise self.ice()

    def proc_list(self, node: py_ast.List) -> ast.ListVal:
        """Process python node.

        class List(expr):
            elts: list[expr]
            ctx: expr_context
        """
        elts = [self.convert(elt) for elt in node.elts]
        valid_elts = [elt for elt in elts if isinstance(elt, ast.Expr)]
        if len(valid_elts) != len(elts):
            raise self.ice("Length mismatch in list elements")
        l_square = self.operator(Tok.LSQUARE, "[")
        r_square = self.operator(Tok.RSQUARE, "]")
        return ast.ListVal(
            values=(
                ast.SubNodeList[ast.Expr](
                    items=valid_elts, delim=Tok.COMMA, kid=valid_elts
                )
                if valid_elts
                else None
            ),
            kid=[*valid_elts] if valid_elts else [l_square, r_square],
        )

    def proc_list_comp(self, node: py_ast.ListComp) -> ast.ListCompr:
        """Process python node.

        class ListComp(expr):
            elt: expr
            generators: list[comprehension]
        """
        elt = self.convert(node.elt)
        generators = [self.convert(gen) for gen in node.generators]
        valid = [gen for gen in generators if isinstance(gen, ast.InnerCompr)]
        if len(generators) != len(valid):
            raise self.ice("Length mismatch in list comp generators")
        compr = ast.SubNodeList[ast.InnerCompr](items=valid, delim=Tok.COMMA, kid=valid)
        if isinstance(elt, ast.Expr):
            return ast.ListCompr(out_expr=elt, compr=valid, kid=[elt, compr])
        else:
            raise self.ice()

    def proc_match(self, node: py_ast.Match) -> ast.MatchStmt:
        """Process python node.

        class Match(stmt):
            subject: expr
            cases: list[match_case]
        """
        subject = self.convert(node.subject)
        cases = [self.convert(i) for i in node.cases]
        valid = [case for case in cases if isinstance(case, ast.MatchCase)]
        if isinstance(subject, ast.Expr):
            return ast.MatchStmt(target=subject, cases=valid, kid=[subject, *valid])
        else:
            raise self.ice()

    def proc_match_as(self, node: py_ast.MatchAs) -> ast.MatchAs | ast.MatchWild:
        """Process python node.

        class MatchAs(pattern):
            pattern: _Pattern | None
            name: _Identifier | None
        """
        pattern = self.convert(node.pattern) if node.pattern else None
        name = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.name if node.name else "_",
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=(
                (node.col_offset + len(node.name))
                if node.name
                else (node.col_offset + 1)
            ),
            pos_start=0,
            pos_end=0,
        )

        if isinstance(pattern, ast.MatchPattern):
            return ast.MatchAs(
                name=name,
                pattern=pattern,
                kid=[name, pattern] if pattern else [name],
            )
        else:
            return ast.MatchWild(kid=[name])

    def proc_match_class(self, node: py_ast.MatchClass) -> ast.MatchArch:
        """Process python node.

        class MatchClass(pattern):
            cls: expr
            patterns: list[pattern]
            kwd_attrs: list[_Identifier]
            kwd_patterns: list[pattern]
        """
        cls = self.convert(node.cls)
        kid = [cls]
        if len(node.patterns) != 0:
            patterns = [self.convert(i) for i in node.patterns]
            valid_patterns = [i for i in patterns if isinstance(i, ast.MatchPattern)]
            if len(patterns) == len(valid_patterns):
                patterns_sub = ast.SubNodeList[ast.MatchPattern](
                    items=valid_patterns, delim=Tok.COMMA, kid=valid_patterns
                )
                kid.append(patterns_sub)
            else:
                raise self.ice()
        else:
            patterns_sub = None

        if len(node.kwd_patterns):
            names: list[ast.Name] = []
            kv_pairs: list[ast.MatchKVPair] = []
            for kwd_attrs in node.kwd_attrs:
                names.append(
                    ast.Name(
                        file_path=self.mod_path,
                        name=Tok.NAME,
                        value=kwd_attrs,
                        line=node.lineno,
                        end_line=node.end_lineno if node.end_lineno else node.lineno,
                        col_start=node.col_offset,
                        col_end=node.col_offset + len(kwd_attrs),
                        pos_start=0,
                        pos_end=0,
                    )
                )
            kwd_patterns = [self.convert(i) for i in node.kwd_patterns]
            valid_kwd_patterns = [
                i for i in kwd_patterns if isinstance(i, ast.MatchPattern)
            ]
            for i in range(len(kwd_patterns)):
                kv_pairs.append(
                    ast.MatchKVPair(
                        key=names[i],
                        value=valid_kwd_patterns[i],
                        kid=[names[i], valid_kwd_patterns[i]],
                    )
                )
            kw_patterns = ast.SubNodeList[ast.MatchKVPair](
                items=kv_pairs, delim=Tok.COMMA, kid=kv_pairs
            )
            kid.append(kw_patterns)
        else:
            kw_patterns = None
        if isinstance(cls, (ast.NameAtom, ast.AtomTrailer)):
            return ast.MatchArch(
                name=cls, arg_patterns=patterns_sub, kw_patterns=kw_patterns, kid=kid
            )
        else:
            raise self.ice()

    def proc_match_mapping(self, node: py_ast.MatchMapping) -> ast.MatchMapping:
        """Process python node.

        class MatchMapping(pattern):
            keys: list[expr]
            patterns: list[pattern]
            rest: _Identifier | None
        """
        values: list[ast.MatchKVPair | ast.MatchStar] = []
        keys = [self.convert(i) for i in node.keys]
        valid_keys = [
            i for i in keys if isinstance(i, (ast.MatchPattern, ast.NameAtom))
        ]
        patterns = [self.convert(i) for i in node.patterns]
        valid_patterns = [i for i in patterns if isinstance(i, ast.MatchPattern)]
        for i in range(len(valid_keys)):
            kv_pair = ast.MatchKVPair(
                key=valid_keys[i],
                value=valid_patterns[i],
                kid=[valid_keys[i], valid_patterns[i]],
            )
            values.append(kv_pair)
        if node.rest:
            name = ast.Name(
                file_path=self.mod_path,
                name=Tok.NAME,
                value=node.rest,
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + len(node.rest),
                pos_start=0,
                pos_end=0,
            )
            values.append(ast.MatchStar(name=name, is_list=True, kid=[name]))
        return ast.MatchMapping(values=values, kid=values)

    def proc_match_or(self, node: py_ast.MatchOr) -> ast.MatchOr:
        """Process python node.

        class MatchOr(pattern):
            patterns: list[pattern]
        """
        patterns = [self.convert(i) for i in node.patterns]
        valid = [i for i in patterns if isinstance(i, ast.MatchPattern)]
        return ast.MatchOr(patterns=valid, kid=valid)

    def proc_match_sequence(self, node: py_ast.MatchSequence) -> ast.MatchSequence:
        """Process python node.

        class MatchSequence(pattern):
            patterns: list[pattern]
        """
        patterns = [self.convert(i) for i in node.patterns]
        valid = [i for i in patterns if isinstance(i, ast.MatchPattern)]
        if len(patterns) == len(valid):
            return ast.MatchSequence(values=valid, kid=valid)
        else:
            raise self.ice()

    def proc_match_singleton(self, node: py_ast.MatchSingleton) -> ast.MatchSingleton:
        """Process python node.

        class MatchSingleton(pattern):
            value: Literal[True, False] | None
        """
        type = Tok.NULL if node.value is None else Tok.BOOL
        ret_type = ast.Null if node.value is None else ast.Bool
        value = ret_type(
            file_path=self.mod_path,
            name=type,
            value=str(node.value),
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(str(node.value)),
            pos_start=0,
            pos_end=0,
        )
        if isinstance(value, (ast.Bool, ast.Null)):
            return ast.MatchSingleton(value=value, kid=[value])
        else:
            raise self.ice()

    def proc_match_star(self, node: py_ast.MatchStar) -> ast.MatchStar:
        """Process python node.

        class MatchStar(pattern):
            name: _Identifier | None
        """
        name = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.name if node.name else "_",
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.name if node.name else "_"),
            pos_start=0,
            pos_end=0,
        )
        return ast.MatchStar(name=name, is_list=True, kid=[name])

    def proc_match_value(self, node: py_ast.MatchValue) -> ast.MatchValue:
        """Process python node.

        class MatchValue(pattern):
            value: expr
        """
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            return ast.MatchValue(value=value, kid=[value])
        else:
            raise self.ice()

    def proc_name(self, node: py_ast.Name) -> ast.Name:
        """Process python node.

        class Name(expr):
        if sys.version_info >= (3, 10):
            __match_args__ = ("id", "ctx")
        id: _Identifier
        ctx: expr_context
        """
        ret = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.id if node.id != "root" else "root_",  # reserved word
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.id),
            pos_start=0,
            pos_end=0,
        )
        return ret

    def proc_named_expr(self, node: py_ast.NamedExpr) -> ast.BinaryExpr:
        """Process python node.

        class NamedExpr(expr):
            target: Name
            value: expr
        """
        target = self.convert(node.target)
        value = self.convert(node.value)
        if isinstance(value, ast.Expr) and isinstance(target, ast.Name):
            return ast.BinaryExpr(
                left=target,
                op=self.operator(Tok.WALRUS_EQ, ":="),
                right=value,
                kid=[target, value],
            )
        else:
            raise self.ice()

    def proc_nonlocal(self, node: py_ast.Nonlocal) -> ast.NonLocalStmt:
        """Process python node.

        class Nonlocal(stmt):
            names: list[_Identifier]
        """
        names: list[ast.NameAtom] = []
        for name in node.names:
            names.append(
                ast.Name(
                    file_path=self.mod_path,
                    name=Tok.NAME,
                    value=name if name != "root" else "root_",
                    line=node.lineno,
                    end_line=node.end_lineno if node.end_lineno else node.lineno,
                    col_start=node.col_offset,
                    col_end=node.col_offset + len(name),
                    pos_start=0,
                    pos_end=0,
                )
            )
        target = ast.SubNodeList[ast.NameAtom](items=names, delim=Tok.COMMA, kid=names)
        return ast.NonLocalStmt(target=target, kid=names)

    def proc_pass(self, node: py_ast.Pass) -> ast.Semi:
        """Process python node."""
        return ast.Semi(
            file_path=self.mod_path,
            name=Tok.SEMI,
            value=";",
            line=0,
            end_line=0,
            col_start=0,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )

    def proc_set(self, node: py_ast.Set) -> ast.SetVal:
        """Process python node.

        class Set(expr):
            elts: list[expr]
        """
        if len(node.elts) != 0:
            elts = [self.convert(i) for i in node.elts]
            valid = [i for i in elts if isinstance(i, (ast.Expr))]
            if len(valid) != len(elts):
                raise self.ice("Length mismatch in set body")
            valid_elts = ast.SubNodeList[ast.Expr](
                items=valid, delim=Tok.COMMA, kid=valid
            )
            kid: list[ast.AstNode] = [*valid]
        else:
            valid_elts = None
            l_brace = self.operator(Tok.LBRACE, "{")
            r_brace = self.operator(Tok.RBRACE, "}")
            kid = [l_brace, r_brace]
        return ast.SetVal(values=valid_elts, kid=kid)

    def proc_set_comp(self, node: py_ast.SetComp) -> ast.ListCompr:
        """Process python node.

        class SetComp(expr):
            elt: expr
            generators: list[comprehension]
        """
        elt = self.convert(node.elt)
        generators = [self.convert(gen) for gen in node.generators]
        valid = [gen for gen in generators if isinstance(gen, ast.InnerCompr)]
        if len(generators) != len(valid):
            raise self.ice("Length mismatch in list comp generators")
        compr = ast.SubNodeList[ast.InnerCompr](items=valid, delim=Tok.COMMA, kid=valid)
        if isinstance(elt, ast.Expr):
            return ast.SetCompr(out_expr=elt, compr=valid, kid=[elt, compr])
        else:
            raise self.ice()

    def proc_slice(self, node: py_ast.Slice) -> ast.IndexSlice:
        """Process python node.

        class Slice(_Slice):
            lower: expr | None
            upper: expr | None
            step: expr | None
        """
        lower = self.convert(node.lower) if node.lower else None
        upper = self.convert(node.upper) if node.upper else None
        step = self.convert(node.step) if node.step else None
        valid_kid = [i for i in [lower, upper, step] if i]
        if not valid_kid:
            valid_kid = [self.operator(Tok.COLON, ":")]
        if (
            (isinstance(lower, ast.Expr) or lower is None)
            and (isinstance(upper, ast.Expr) or upper is None)
            and (isinstance(step, ast.Expr) or step is None)
        ):
            return ast.IndexSlice(
                start=lower,
                stop=upper,
                step=step,
                is_range=True,
                kid=valid_kid,
            )
        else:
            raise self.ice()

    def proc_starred(self, node: py_ast.Starred) -> ast.UnaryExpr:
        """Process python node.

        class Starred(expr):
            value: expr
            ctx: expr_context
        """
        star_tok = self.operator(Tok.STAR_MUL, "*")
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            return ast.UnaryExpr(operand=value, op=star_tok, kid=[value, star_tok])
        else:
            raise self.ice()

    def proc_subscript(self, node: py_ast.Subscript) -> ast.AtomTrailer:
        """Process python node.

        class Subscript(expr):
            value: expr
            slice: _Slice
            ctx: expr_context
        """
        value = self.convert(node.value)
        slice = self.convert(node.slice)
        if not isinstance(slice, ast.IndexSlice) and isinstance(slice, ast.Expr):
            slice = ast.IndexSlice(
                start=slice,
                stop=None,
                step=None,
                is_range=False,
                kid=[slice],
            )
        if isinstance(value, ast.Expr) and isinstance(slice, ast.IndexSlice):
            return ast.AtomTrailer(
                target=value,
                right=slice,
                is_attr=False,
                is_null_ok=False,
                kid=[value, slice],
            )
        else:
            raise self.ice()

    def proc_try(self, node: py_ast.Try | py_ast.TryStar) -> ast.TryStmt:
        """Process python node.

        class Try(stmt):
            body: list[stmt]
            handlers: list[ExceptHandler]
            orelse: list[stmt]
            finalbody: list[stmt]
        """
        body = [self.convert(i) for i in node.body]
        valid = [i for i in body if isinstance(i, (ast.CodeBlockStmt))]
        if len(valid) != len(body):
            raise self.ice("Length mismatch in try body")
        valid_body = ast.SubNodeList[ast.CodeBlockStmt](
            items=valid,
            delim=Tok.WS,
            kid=valid,
            left_enc=self.operator(Tok.LBRACE, "{"),
            right_enc=self.operator(Tok.RBRACE, "}"),
        )
        kid: list[ast.AstNode] = [valid_body]

        if len(node.handlers) != 0:
            handlers = [self.convert(i) for i in node.handlers]
            valid_handlers = [i for i in handlers if isinstance(i, (ast.Except))]
            if len(handlers) != len(valid_handlers):
                raise self.ice("Length mismatch in try handlers")
            excepts = ast.SubNodeList[ast.Except](
                items=valid_handlers, delim=Tok.WS, kid=valid_handlers
            )
            kid.append(excepts)
        else:
            excepts = None

        if len(node.orelse) != 0:
            orelse = [self.convert(i) for i in node.orelse]
            valid_orelse = [i for i in orelse if isinstance(i, (ast.CodeBlockStmt))]
            if len(orelse) != len(valid_orelse):
                raise self.ice("Length mismatch in try orelse")
            else_body = ast.SubNodeList[ast.CodeBlockStmt](
                items=valid_orelse,
                delim=Tok.WS,
                kid=valid_orelse,
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
            elsestmt = ast.ElseStmt(body=else_body, kid=[else_body])
            kid.append(else_body)
        else:
            else_body = None

        if len(node.finalbody) != 0:
            finalbody = [self.convert(i) for i in node.finalbody]
            valid_finalbody = [
                i for i in finalbody if isinstance(i, (ast.CodeBlockStmt))
            ]
            if len(finalbody) != len(valid_finalbody):
                raise self.ice("Length mismatch in try finalbody")
            finally_body = ast.SubNodeList[ast.CodeBlockStmt](
                items=valid_finalbody,
                delim=Tok.WS,
                kid=valid_finalbody,
                left_enc=self.operator(Tok.LBRACE, "{"),
                right_enc=self.operator(Tok.RBRACE, "}"),
            )
            finally_stmt = ast.FinallyStmt(body=finally_body, kid=[finally_body])

            kid.append(finally_stmt)
        else:
            finally_body = None
        ret = ast.TryStmt(
            body=valid_body,
            excepts=excepts,
            else_body=elsestmt if else_body else None,
            finally_body=finally_stmt if finally_body else None,
            kid=kid,
        )
        return ret

    def proc_try_star(self, node: py_ast.TryStar) -> ast.TryStmt:
        """Process python node.

        class Try(stmt):
            body: list[stmt]
            handlers: list[ExceptHandler]
            orelse: list[stmt]
            finalbody: list[stmt]
        """
        return self.proc_try(node)

    def proc_tuple(self, node: py_ast.Tuple) -> ast.TupleVal:
        """Process python node.

        class Tuple(expr):
            elts: list[expr]
            ctx: expr_context
        """
        elts = [self.convert(elt) for elt in node.elts]
        if len(node.elts) != 0:
            valid = [i for i in elts if isinstance(i, (ast.Expr, ast.KWPair))]
            if len(elts) != len(valid):
                raise self.ice("Length mismatch in tuple elts")
            valid_elts = ast.SubNodeList[ast.Expr | ast.KWPair](
                items=valid, delim=Tok.COMMA, kid=valid
            )
            kid = elts
        else:
            l_paren = self.operator(Tok.LPAREN, "(")
            r_paren = self.operator(Tok.RPAREN, ")")
            valid_elts = None
            kid = [l_paren, r_paren]
        return ast.TupleVal(values=valid_elts, kid=kid)

    def proc_yield(self, node: py_ast.Yield) -> ast.YieldExpr:
        """Process python node.

        class Yield(expr):
            value: expr | None
        """
        value = self.convert(node.value) if node.value else None
        if isinstance(value, ast.Expr):
            return ast.YieldExpr(expr=value, with_from=False, kid=[value])
        elif not value:
            return ast.YieldExpr(
                expr=None, with_from=False, kid=[self.operator(Tok.KW_YIELD, "yield")]
            )
        else:
            raise self.ice()

    def proc_yield_from(self, node: py_ast.YieldFrom) -> ast.YieldExpr:
        """Process python node."""
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            return ast.YieldExpr(expr=value, with_from=True, kid=[value])
        else:
            raise self.ice()

    def proc_alias(self, node: py_ast.alias) -> ast.ExprAsItem:
        """Process python node.

        class alias(AST):
            name: _Identifier
            asname: _Identifier | None
        """
        name = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.name,
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.name),
            pos_start=0,
            pos_end=0,
        )
        asname = (
            ast.Name(
                file_path=self.mod_path,
                name=Tok.NAME,
                value=node.asname,
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + len(node.asname),
                pos_start=0,
                pos_end=0,
            )
            if node.asname
            else None
        )
        return ast.ExprAsItem(
            expr=name, alias=asname, kid=[name, asname] if asname else [name]
        )

    def proc_arg(self, node: py_ast.arg) -> ast.ParamVar:
        """Process python node.

        class arg(AST):
            arg: _Identifier
            annotation: expr | None
        """
        name = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.arg if node.arg != "root" else "root_",  # reserved word
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.arg),
            pos_start=0,
            pos_end=0,
        )
        ann_expr = (
            self.convert(node.annotation)
            if node.annotation
            else ast.Name(
                file_path=self.mod_path,
                name=Tok.NAME,
                value="Any",
                line=node.lineno,
                end_line=node.end_lineno if node.end_lineno else node.lineno,
                col_start=node.col_offset,
                col_end=node.col_offset + 3,
                pos_start=0,
                pos_end=0,
            )
        )
        if not isinstance(ann_expr, ast.Expr):
            raise self.ice("Expected annotation to be an expression")
        annot = ast.SubTag[ast.Expr](tag=ann_expr, kid=[ann_expr])
        paramvar = ast.ParamVar(
            name=name, type_tag=annot, unpack=None, value=None, kid=[name, annot]
        )
        return paramvar

    def proc_arguments(self, node: py_ast.arguments) -> ast.FuncSignature:
        """Process python node.

        class arguments(AST):
            args: list[arg]
            vararg: arg | None
            kwonlyargs: list[arg]
            kw_defaults: list[expr | None]
            kwarg: arg | None
            defaults: list[expr]
        """
        args = [self.convert(arg) for arg in node.args]
        vararg = self.convert(node.vararg) if node.vararg else None
        if vararg and isinstance(vararg, ast.ParamVar):
            vararg.unpack = ast.Token(
                file_path=self.mod_path,
                name=Tok.STAR_MUL,
                value="*",
                line=vararg.loc.first_line,
                end_line=vararg.loc.last_line,
                col_start=vararg.loc.col_start,
                col_end=vararg.loc.col_end,
                pos_start=0,
                pos_end=0,
            )
            vararg.add_kids_left([vararg.unpack])
        kwonlyargs = [self.convert(arg) for arg in node.kwonlyargs]
        for i in range(len(kwonlyargs)):
            kwa = kwonlyargs[i]
            kwd = node.kw_defaults[i]
            kwdefault = self.convert(kwd) if kwd else None
            if (
                kwdefault
                and isinstance(kwa, ast.ParamVar)
                and isinstance(kwdefault, ast.Expr)
            ):
                kwa.value = kwdefault
                kwa.add_kids_right([kwa.value])
        kwarg = self.convert(node.kwarg) if node.kwarg else None
        if kwarg and isinstance(kwarg, ast.ParamVar):
            kwarg.unpack = ast.Token(
                file_path=self.mod_path,
                name=Tok.STAR_POW,
                value="**",
                line=kwarg.loc.first_line,
                end_line=kwarg.loc.last_line,
                col_start=kwarg.loc.col_start,
                col_end=kwarg.loc.col_end,
                pos_start=0,
                pos_end=0,
            )
            kwarg.add_kids_left([kwarg.unpack])
        defaults = [self.convert(expr) for expr in node.defaults]
        params = [*args]
        for param, default in zip(params[::-1], defaults[::-1]):
            if isinstance(default, ast.Expr) and isinstance(param, ast.ParamVar):
                param.value = default
                param.add_kids_right([default])
        if vararg:
            params.append(vararg)
        params += kwonlyargs
        if kwarg:
            params.append(kwarg)
        params += defaults

        valid_params = [param for param in params if isinstance(param, ast.ParamVar)]
        if valid_params:
            fs_params = ast.SubNodeList[ast.ParamVar](
                items=valid_params, delim=Tok.COMMA, kid=valid_params
            )
            return ast.FuncSignature(
                params=fs_params,
                return_type=None,
                kid=[fs_params],
            )
        else:
            _lparen = self.operator(Tok.LPAREN, "(")
            _rparen = self.operator(Tok.RPAREN, ")")
            return ast.FuncSignature(
                params=None,
                return_type=None,
                kid=[_lparen, _rparen],
            )

    def operator(self, tok: Tok, value: str) -> ast.Token:
        """Create an operator token."""
        return ast.Token(
            file_path=self.mod_path,
            name=tok,
            value=value,
            line=0,
            end_line=0,
            col_start=0,
            col_end=0,
            pos_start=0,
            pos_end=0,
        )

    def proc_and(self, node: py_ast.And) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.KW_AND, "and")

    def proc_or(self, node: py_ast.Or) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.KW_OR, "or")

    def proc_add(self, node: py_ast.Add) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.PLUS, "+")

    def proc_bit_and(self, node: py_ast.BitAnd) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.BW_AND, "&")

    def proc_bit_or(self, node: py_ast.BitOr) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.BW_OR, "|")

    def proc_bit_xor(self, node: py_ast.BitXor) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.BW_XOR, "^")

    def proc_div(self, node: py_ast.Div) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.DIV, "/")

    def proc_floor_div(self, node: py_ast.FloorDiv) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.FLOOR_DIV, "//")

    def proc_l_shift(self, node: py_ast.LShift) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.LSHIFT, "<<")

    def proc_mod(self, node: py_ast.Mod) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.MOD, "%")

    def proc_mult(self, node: py_ast.Mult) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.STAR_MUL, "*")

    def proc_mat_mult(self, node: py_ast.MatMult) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.DECOR_OP, "@")

    def proc_pow(self, node: py_ast.Pow) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.STAR_POW, "**")

    def proc_r_shift(self, node: py_ast.RShift) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.RSHIFT, ">>")

    def proc_sub(self, node: py_ast.Sub) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.MINUS, "-")

    def proc_invert(self, node: py_ast.Invert) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.BW_NOT, "~")

    def proc_not(self, node: py_ast.Not) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.NOT, "not")

    def proc_u_add(self, node: py_ast.UAdd) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.PLUS, "+")

    def proc_u_sub(self, node: py_ast.USub) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.MINUS, "-")

    def proc_eq(self, node: py_ast.Eq) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.EE, "==")

    def proc_gt(self, node: py_ast.Gt) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.GT, ">")

    def proc_gt_e(self, node: py_ast.GtE) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.GTE, ">=")

    def proc_in(self, node: py_ast.In) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.KW_IN, "in")

    def proc_is(self, node: py_ast.Is) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.KW_IS, "is")

    def proc_is_not(self, node: py_ast.IsNot) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.KW_ISN, "is not")

    def proc_lt(self, node: py_ast.Lt) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.LT, "<")

    def proc_lt_e(self, node: py_ast.LtE) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.LTE, "<=")

    def proc_not_eq(self, node: py_ast.NotEq) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.NE, "!=")

    def proc_not_in(self, node: py_ast.NotIn) -> ast.Token:
        """Process python node."""
        return self.operator(Tok.KW_NIN, "not in")

    def proc_comprehension(self, node: py_ast.comprehension) -> ast.InnerCompr:
        """Process python node.

        class comprehension(AST):
            target: expr
            iter: expr
            ifs: list[expr]
            is_async: int
        """
        target = self.convert(node.target)
        iter = self.convert(node.iter)
        if len(node.ifs) != 0:
            ifs_list = [self.convert(ifs) for ifs in node.ifs]
            valid = [ifs for ifs in ifs_list if isinstance(ifs, ast.Expr)]
        else:
            valid = None
        is_async = node.is_async > 0
        if isinstance(target, ast.Expr) and isinstance(iter, ast.Expr):
            return ast.InnerCompr(
                is_async=is_async,
                target=target,
                collection=iter,
                conditional=valid,
                kid=[target, iter, *valid] if valid else [target, iter],
            )
        else:
            raise self.ice()

    def proc_keyword(self, node: py_ast.keyword) -> ast.KWPair:
        """Process python node.

        class keyword(AST):
        if sys.version_info >= (3, 10):
            __match_args__ = ("arg", "value")
        arg: _Identifier | None
        value: expr
        """
        arg = ast.Name(
            file_path=self.mod_path,
            name=Tok.NAME,
            value=node.arg if node.arg else "_",
            line=node.lineno,
            end_line=node.end_lineno if node.end_lineno else node.lineno,
            col_start=node.col_offset,
            col_end=node.col_offset + len(node.arg if node.arg else "_"),
            pos_start=0,
            pos_end=0,
        )
        value = self.convert(node.value)
        if isinstance(value, ast.Expr):
            return ast.KWPair(key=arg, value=value, kid=[arg, value])
        else:
            raise self.ice()

    def proc_match_case(self, node: py_ast.match_case) -> ast.MatchCase:
        """Process python node.

        class match_case(AST):
            pattern: _Pattern
            guard: expr | None
            body: list[stmt]
        """
        pattern = self.convert(node.pattern)
        guard = self.convert(node.guard) if node.guard else None
        body = [self.convert(i) for i in node.body]
        valid = [i for i in body if isinstance(i, ast.CodeBlockStmt)]
        if isinstance(pattern, ast.MatchPattern) and (
            isinstance(guard, ast.Expr) or guard is None
        ):
            return ast.MatchCase(
                pattern=pattern,
                guard=guard,
                body=valid,
                kid=([pattern, guard, *valid] if guard else [pattern, *valid]),
            )
        else:
            raise self.ice()

    def proc_withitem(self, node: py_ast.withitem) -> ast.ExprAsItem:
        """Process python node.

        class withitem(AST):
            context_expr: expr
            optional_vars: expr | None
        """
        context_expr = self.convert(node.context_expr)
        optional_vars = self.convert(node.optional_vars) if node.optional_vars else None
        if isinstance(context_expr, ast.Expr) and (
            isinstance(optional_vars, ast.Expr) or optional_vars is None
        ):
            return ast.ExprAsItem(
                expr=context_expr,
                alias=optional_vars if optional_vars else None,
                kid=[context_expr, optional_vars] if optional_vars else [context_expr],
            )
        else:
            raise self.ice()

    def proc_param_spec(self, node: py_ast.ParamSpec) -> None:
        """Process python node."""

    def proc_type_alias(self, node: py_ast.TypeAlias) -> None:
        """Process python node."""

    def proc_type_var(self, node: py_ast.TypeVar) -> None:
        """Process python node."""

    def proc_type_var_tuple(self, node: py_ast.TypeVarTuple) -> None:
        """Process python node."""

    def convert_to_doc(self, string: ast.String) -> None:
        """Convert a string to a docstring."""
        string.value = f'""{string.value}""'

    def aug_op_map(self, tok_dict: dict, op: ast.Token) -> str:
        """aug_mapper."""
        op.value += "="
        for _key, value in tok_dict.items():
            if value == op.value:
                break
        return _key



"""Jac Language Features."""

from __future__ import annotations

import types
from typing import Any, Callable, Optional, Type, TypeAlias, Union

from jaclang.compiler.absyntree import Module
from jaclang.core.constructs import (
    Architype,
    EdgeArchitype,
    Memory,
    NodeArchitype,
    Root,
    WalkerArchitype,
)
from jaclang.plugin.default import ExecutionContext
from jaclang.plugin.spec import JacBuiltin, JacCmdSpec, JacFeatureSpec, T


import pluggy

pm = pluggy.PluginManager("jac")
pm.add_hookspecs(JacFeatureSpec)
pm.add_hookspecs(JacCmdSpec)
pm.add_hookspecs(JacBuiltin)


class JacFeature:
    """Jac Feature."""

    import abc
    from jaclang.compiler.constant import EdgeDir
    from jaclang.core.constructs import DSFunc

    RootType: TypeAlias = Root
    Obj: TypeAlias = Architype
    Node: TypeAlias = NodeArchitype
    Edge: TypeAlias = EdgeArchitype
    Walker: TypeAlias = WalkerArchitype

    @staticmethod
    def context(session: str = "") -> ExecutionContext:
        """Create execution context."""
        return pm.hook.context(session=session)

    @staticmethod
    def reset_context() -> None:
        """Reset execution context."""
        return pm.hook.reset_context()

    @staticmethod
    def memory_hook() -> Memory | None:
        """Create memory abstraction."""
        return pm.hook.memory_hook()

    @staticmethod
    def make_architype(
        cls: type,
        arch_base: Type[Architype],
        on_entry: list[DSFunc],
        on_exit: list[DSFunc],
    ) -> Type[Architype]:
        """Create a obj architype."""
        return pm.hook.make_architype(
            cls=cls, on_entry=on_entry, on_exit=on_exit, arch_base=arch_base
        )

    @staticmethod
    def make_obj(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a obj architype."""
        return pm.hook.make_obj(on_entry=on_entry, on_exit=on_exit)

    @staticmethod
    def make_node(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a node architype."""
        return pm.hook.make_node(on_entry=on_entry, on_exit=on_exit)

    @staticmethod
    def make_edge(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a edge architype."""
        return pm.hook.make_edge(on_entry=on_entry, on_exit=on_exit)

    @staticmethod
    def make_walker(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a walker architype."""
        return pm.hook.make_walker(on_entry=on_entry, on_exit=on_exit)

    @staticmethod
    def jac_import(
        target: str,
        base_path: str,
        absorb: bool = False,
        cachable: bool = True,
        mdl_alias: Optional[str] = None,
        override_name: Optional[str] = None,
        mod_bundle: Optional[Module | str] = None,
        lng: Optional[str] = "jac",
        items: Optional[dict[str, Union[str, bool]]] = None,
    ) -> Optional[types.ModuleType]:
        """Core Import Process."""
        return pm.hook.jac_import(
            target=target,
            base_path=base_path,
            absorb=absorb,
            cachable=cachable,
            mdl_alias=mdl_alias,
            override_name=override_name,
            mod_bundle=mod_bundle,
            lng=lng,
            items=items,
        )

    @staticmethod
    def create_test(test_fun: Callable) -> Callable:
        """Create a test."""
        return pm.hook.create_test(test_fun=test_fun)

    @staticmethod
    def run_test(
        filepath: str,
        filter: Optional[str] = None,
        xit: bool = False,
        maxfail: Optional[int] = None,
        directory: Optional[str] = None,
        verbose: bool = False,
    ) -> bool:
        """Run the test suite in the specified .jac file."""
        return pm.hook.run_test(
            filepath=filepath,
            filter=filter,
            xit=xit,
            maxfail=maxfail,
            directory=directory,
            verbose=verbose,
        )

    @staticmethod
    def elvis(op1: Optional[T], op2: T) -> T:
        """Jac's elvis operator feature."""
        return pm.hook.elvis(op1=op1, op2=op2)

    @staticmethod
    def has_instance_default(gen_func: Callable[[], T]) -> T:
        """Jac's has container default feature."""
        return pm.hook.has_instance_default(gen_func=gen_func)

    @staticmethod
    def spawn_call(op1: Architype, op2: Architype) -> WalkerArchitype:
        """Jac's spawn operator feature."""
        return pm.hook.spawn_call(op1=op1, op2=op2)

    @staticmethod
    def report(expr: Any) -> Any:  # noqa: ANN401
        """Jac's report stmt feature."""
        return pm.hook.report(expr=expr)

    @staticmethod
    def ignore(
        walker: WalkerArchitype,
        expr: (
            list[NodeArchitype | EdgeArchitype]
            | list[NodeArchitype]
            | list[EdgeArchitype]
            | NodeArchitype
            | EdgeArchitype
        ),
    ) -> bool:  # noqa: ANN401
        """Jac's ignore stmt feature."""
        return pm.hook.ignore(walker=walker, expr=expr)

    @staticmethod
    def visit_node(
        walker: WalkerArchitype,
        expr: (
            list[NodeArchitype | EdgeArchitype]
            | list[NodeArchitype]
            | list[EdgeArchitype]
            | NodeArchitype
            | EdgeArchitype
        ),
    ) -> bool:  # noqa: ANN401
        """Jac's visit stmt feature."""
        return pm.hook.visit_node(walker=walker, expr=expr)

    @staticmethod
    def disengage(walker: WalkerArchitype) -> bool:  # noqa: ANN401
        """Jac's disengage stmt feature."""
        return pm.hook.disengage(walker=walker)

    @staticmethod
    def edge_ref(
        node_obj: NodeArchitype | list[NodeArchitype],
        target_obj: Optional[NodeArchitype | list[NodeArchitype]],
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
        edges_only: bool = False,
    ) -> list[NodeArchitype] | list[EdgeArchitype]:
        """Jac's apply_dir stmt feature."""
        return pm.hook.edge_ref(
            node_obj=node_obj,
            target_obj=target_obj,
            dir=dir,
            filter_func=filter_func,
            edges_only=edges_only,
        )

    @staticmethod
    def connect(
        left: NodeArchitype | list[NodeArchitype],
        right: NodeArchitype | list[NodeArchitype],
        edge_spec: Callable[[], EdgeArchitype],
        edges_only: bool = False,
    ) -> list[NodeArchitype] | list[EdgeArchitype]:
        """Jac's connect operator feature.

        Note: connect needs to call assign compr with tuple in op
        """
        return pm.hook.connect(
            left=left, right=right, edge_spec=edge_spec, edges_only=edges_only
        )

    @staticmethod
    def disconnect(
        left: NodeArchitype | list[NodeArchitype],
        right: NodeArchitype | list[NodeArchitype],
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
    ) -> bool:
        """Jac's disconnect operator feature."""
        return pm.hook.disconnect(
            left=left,
            right=right,
            dir=dir,
            filter_func=filter_func,
        )

    @staticmethod
    def assign_compr(
        target: list[T], attr_val: tuple[tuple[str], tuple[Any]]
    ) -> list[T]:
        """Jac's assign comprehension feature."""
        return pm.hook.assign_compr(target=target, attr_val=attr_val)

    @staticmethod
    def get_root() -> Root:
        """Jac's root getter."""
        return pm.hook.get_root()

    @staticmethod
    def get_root_type() -> Type[Root]:
        """Jac's root type getter."""
        return pm.hook.get_root_type()

    @staticmethod
    def build_edge(
        is_undirected: bool,
        conn_type: Optional[Type[EdgeArchitype] | EdgeArchitype],
        conn_assign: Optional[tuple[tuple, tuple]],
    ) -> Callable[[], EdgeArchitype]:
        """Jac's root getter."""
        return pm.hook.build_edge(
            is_undirected=is_undirected, conn_type=conn_type, conn_assign=conn_assign
        )

    @staticmethod
    def get_semstr_type(
        file_loc: str, scope: str, attr: str, return_semstr: bool
    ) -> Optional[str]:
        """Jac's get_semstr_type feature."""
        return pm.hook.get_semstr_type(
            file_loc=file_loc, scope=scope, attr=attr, return_semstr=return_semstr
        )

    @staticmethod
    def obj_scope(file_loc: str, attr: str) -> str:
        """Jac's get_semstr_type feature."""
        return pm.hook.obj_scope(file_loc=file_loc, attr=attr)

    @staticmethod
    def get_sem_type(file_loc: str, attr: str) -> tuple[str | None, str | None]:
        """Jac's get_semstr_type feature."""
        return pm.hook.get_sem_type(file_loc=file_loc, attr=attr)

    @staticmethod
    def with_llm(
        file_loc: str,
        model: Any,  # noqa: ANN401
        model_params: dict[str, Any],
        scope: str,
        incl_info: list[tuple[str, str]],
        excl_info: list[tuple[str, str]],
        inputs: list[tuple[str, str, str, Any]],
        outputs: tuple,
        action: str,
    ) -> Any:  # noqa: ANN401
        """Jac's with_llm feature."""
        return pm.hook.with_llm(
            file_loc=file_loc,
            model=model,
            model_params=model_params,
            scope=scope,
            incl_info=incl_info,
            excl_info=excl_info,
            inputs=inputs,
            outputs=outputs,
            action=action,
        )


class JacCmd:
    """Jac CLI command."""

    @staticmethod
    def create_cmd() -> None:
        """Create Jac CLI cmds."""
        return pm.hook.create_cmd()



"""Core constructs for Jac Language."""

from __future__ import annotations

import types
from dataclasses import dataclass, field
from typing import Any, Callable, Optional
from uuid import UUID, uuid4

from jaclang.compiler.constant import EdgeDir
from jaclang.core.utils import collect_node_connections


@dataclass(eq=False)
class ElementAnchor:
    """Element Anchor."""

    obj: Architype
    id: UUID = field(default_factory=uuid4)


@dataclass(eq=False)
class ObjectAnchor(ElementAnchor):
    """Object Anchor."""

    def spawn_call(self, walk: WalkerArchitype) -> WalkerArchitype:
        """Invoke data spatial call."""
        return walk._jac_.spawn_call(self.obj)


@dataclass(eq=False)
class NodeAnchor(ObjectAnchor):
    """Node Anchor."""

    obj: NodeArchitype
    edges: list[EdgeArchitype] = field(default_factory=lambda: [])
    edge_ids: list[UUID] = field(default_factory=lambda: [])
    persistent: bool = False

    def __getstate__(self) -> dict:
        """Override getstate for pickle and shelve."""
        state = self.__dict__.copy()
        state.pop("obj")
        if self.edges and "edges" in state:
            edges = state.pop("edges")
            state["edge_ids"] = [e._jac_.id for e in edges]

        return state

    def __setstate__(self, state: dict) -> None:
        """Override setstate for pickle and shelve."""
        self.__dict__.update(state)
        if "edge_ids" in state:
            self.edge_ids = state.pop("edge_ids")

    def populate_edges(self) -> None:
        """Populate edges from edge ids."""
        from jaclang.plugin.feature import JacFeature as Jac

        if len(self.edges) == 0 and len(self.edge_ids) > 0:
            for e_id in self.edge_ids:
                edge = Jac.context().get_obj(e_id)
                if edge is None:
                    raise ValueError(f"Edge with id {e_id} not found.")
                elif not isinstance(edge, EdgeArchitype):
                    raise ValueError(f"Object with id {e_id} is not an edge.")
                else:
                    self.edges.append(edge)
            self.edge_ids.clear()

    def connect_node(self, nd: NodeArchitype, edg: EdgeArchitype) -> NodeArchitype:
        """Connect a node with given edge."""
        edg._jac_.attach(self.obj, nd)
        return self.obj

    def get_edges(
        self,
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
        target_obj: Optional[list[NodeArchitype]],
    ) -> list[EdgeArchitype]:
        """Get edges connected to this node."""
        self.populate_edges()

        edge_list: list[EdgeArchitype] = [*self.edges]
        ret_edges: list[EdgeArchitype] = []
        edge_list = filter_func(edge_list) if filter_func else edge_list
        for e in edge_list:
            if (
                e._jac_.target
                and e._jac_.source
                and (
                    dir in [EdgeDir.OUT, EdgeDir.ANY]
                    and self.obj == e._jac_.source
                    and (not target_obj or e._jac_.target in target_obj)
                )
                or (
                    dir in [EdgeDir.IN, EdgeDir.ANY]
                    and self.obj == e._jac_.target
                    and (not target_obj or e._jac_.source in target_obj)
                )
            ):
                ret_edges.append(e)
        return ret_edges

    def edges_to_nodes(
        self,
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
        target_obj: Optional[list[NodeArchitype]],
    ) -> list[NodeArchitype]:
        """Get set of nodes connected to this node."""
        self.populate_edges()
        for edge in self.edges:
            edge.populate_nodes()
        edge_list: list[EdgeArchitype] = [*self.edges]
        node_list: list[NodeArchitype] = []
        edge_list = filter_func(edge_list) if filter_func else edge_list
        for e in edge_list:
            if e._jac_.target and e._jac_.source:
                if (
                    dir in [EdgeDir.OUT, EdgeDir.ANY]
                    and self.obj == e._jac_.source
                    and (not target_obj or e._jac_.target in target_obj)
                ):
                    node_list.append(e._jac_.target)
                if (
                    dir in [EdgeDir.IN, EdgeDir.ANY]
                    and self.obj == e._jac_.target
                    and (not target_obj or e._jac_.source in target_obj)
                ):
                    node_list.append(e._jac_.source)
        return node_list

    def gen_dot(self, dot_file: Optional[str] = None) -> str:
        """Generate Dot file for visualizing nodes and edges."""
        visited_nodes: set[NodeAnchor] = set()
        connections: set[tuple[NodeArchitype, NodeArchitype, str]] = set()
        unique_node_id_dict = {}

        collect_node_connections(self, visited_nodes, connections)
        dot_content = 'digraph {\nnode [style="filled", shape="ellipse", fillcolor="invis", fontcolor="black"];\n'
        for idx, i in enumerate([nodes_.obj for nodes_ in visited_nodes]):
            unique_node_id_dict[i] = (i.__class__.__name__, str(idx))
            dot_content += f'{idx} [label="{i}"];\n'
        dot_content += 'edge [color="gray", style="solid"];\n'

        for pair in list(set(connections)):
            dot_content += (
                f"{unique_node_id_dict[pair[0]][1]} -> {unique_node_id_dict[pair[1]][1]}"
                f' [label="{pair[2]}"];\n'
            )
        if dot_file:
            with open(dot_file, "w") as f:
                f.write(dot_content + "}")
        return dot_content + "}"


@dataclass(eq=False)
class EdgeAnchor(ObjectAnchor):
    """Edge Anchor."""

    obj: EdgeArchitype
    source: Optional[NodeArchitype] = None
    target: Optional[NodeArchitype] = None
    source_id: Optional[UUID] = None
    target_id: Optional[UUID] = None
    is_undirected: bool = False
    persistent: bool = False

    def __getstate__(self) -> dict:
        """Override getstate for pickle and shelve."""
        state = self.__dict__.copy()
        state.pop("obj")

        if self.source:
            state["source_id"] = self.source._jac_.id
            state.pop("source")

        if self.target:
            state["target_id"] = self.target._jac_.id
            state.pop("target")

        return state

    def __setstate__(self, state: dict) -> None:
        """Override setstate for pickle and shelve."""
        self.__dict__.update(state)

    def attach(
        self, src: NodeArchitype, trg: NodeArchitype, is_undirected: bool = False
    ) -> EdgeAnchor:
        """Attach edge to nodes."""
        self.source = src
        self.target = trg
        self.is_undirected = is_undirected
        src._jac_.edges.append(self.obj)
        trg._jac_.edges.append(self.obj)
        return self

    def detach(
        self, src: NodeArchitype, trg: NodeArchitype, is_undirected: bool = False
    ) -> None:
        """Detach edge from nodes."""
        self.is_undirected = is_undirected
        src._jac_.edges.remove(self.obj)
        trg._jac_.edges.remove(self.obj)
        self.source = None
        self.target = None
        del self

    def spawn_call(self, walk: WalkerArchitype) -> WalkerArchitype:
        """Invoke data spatial call."""
        if self.target:
            return walk._jac_.spawn_call(self.target)
        else:
            raise ValueError("Edge has no target.")


@dataclass(eq=False)
class WalkerAnchor(ObjectAnchor):
    """Walker Anchor."""

    obj: WalkerArchitype
    path: list[Architype] = field(default_factory=lambda: [])
    next: list[Architype] = field(default_factory=lambda: [])
    ignores: list[Architype] = field(default_factory=lambda: [])
    disengaged: bool = False

    def visit_node(
        self,
        nds: (
            list[NodeArchitype | EdgeArchitype]
            | list[NodeArchitype]
            | list[EdgeArchitype]
            | NodeArchitype
            | EdgeArchitype
        ),
    ) -> bool:
        """Walker visits node."""
        nd_list: list[NodeArchitype | EdgeArchitype]
        if not isinstance(nds, list):
            nd_list = [nds]
        else:
            nd_list = list(nds)
        before_len = len(self.next)
        for i in nd_list:
            if i not in self.ignores:
                if isinstance(i, NodeArchitype):
                    self.next.append(i)
                elif isinstance(i, EdgeArchitype):
                    if i._jac_.target:
                        self.next.append(i._jac_.target)
                    else:
                        raise ValueError("Edge has no target.")
        return len(self.next) > before_len

    def ignore_node(
        self,
        nds: (
            list[NodeArchitype | EdgeArchitype]
            | list[NodeArchitype]
            | list[EdgeArchitype]
            | NodeArchitype
            | EdgeArchitype
        ),
    ) -> bool:
        """Walker ignores node."""
        nd_list: list[NodeArchitype | EdgeArchitype]
        if not isinstance(nds, list):
            nd_list = [nds]
        else:
            nd_list = list(nds)
        before_len = len(self.ignores)
        for i in nd_list:
            if i not in self.ignores:
                if isinstance(i, NodeArchitype):
                    self.ignores.append(i)
                elif isinstance(i, EdgeArchitype):
                    if i._jac_.target:
                        self.ignores.append(i._jac_.target)
                    else:
                        raise ValueError("Edge has no target.")
        return len(self.ignores) > before_len

    def disengage_now(self) -> None:
        """Disengage walker from traversal."""
        self.disengaged = True

    def spawn_call(self, nd: Architype) -> WalkerArchitype:
        """Invoke data spatial call."""
        self.path = []
        self.next = [nd]
        while len(self.next):
            nd = self.next.pop(0)
            for i in nd._jac_entry_funcs_:
                if not i.trigger or isinstance(self.obj, i.trigger):
                    if i.func:
                        i.func(nd, self.obj)
                    else:
                        raise ValueError(f"No function {i.name} to call.")
                if self.disengaged:
                    return self.obj
            for i in self.obj._jac_entry_funcs_:
                if not i.trigger or isinstance(nd, i.trigger):
                    if i.func:
                        i.func(self.obj, nd)
                    else:
                        raise ValueError(f"No function {i.name} to call.")
                if self.disengaged:
                    return self.obj
            for i in self.obj._jac_exit_funcs_:
                if not i.trigger or isinstance(nd, i.trigger):
                    if i.func:
                        i.func(self.obj, nd)
                    else:
                        raise ValueError(f"No function {i.name} to call.")
                if self.disengaged:
                    return self.obj
            for i in nd._jac_exit_funcs_:
                if not i.trigger or isinstance(self.obj, i.trigger):
                    if i.func:
                        i.func(nd, self.obj)
                    else:
                        raise ValueError(f"No function {i.name} to call.")
                if self.disengaged:
                    return self.obj
        self.ignores = []
        return self.obj


class Architype:
    """Architype Protocol."""

    _jac_entry_funcs_: list[DSFunc]
    _jac_exit_funcs_: list[DSFunc]

    def __init__(self) -> None:
        """Create default architype."""
        self._jac_: ObjectAnchor = ObjectAnchor(obj=self)

    def __hash__(self) -> int:
        """Override hash for architype."""
        return hash(self._jac_.id)

    def __eq__(self, other: object) -> bool:
        """Override equality for architype."""
        if not isinstance(other, Architype):
            return False
        else:
            return self._jac_.id == other._jac_.id

    def __repr__(self) -> str:
        """Override repr for architype."""
        return f"{self.__class__.__name__}"

    def __getstate__(self) -> dict:
        """Override getstate for pickle and shelve."""
        raise NotImplementedError


class NodeArchitype(Architype):
    """Node Architype Protocol."""

    _jac_: NodeAnchor

    def __init__(self) -> None:
        """Create node architype."""
        from jaclang.plugin.feature import JacFeature as Jac

        self._jac_: NodeAnchor = NodeAnchor(obj=self)
        Jac.context().save_obj(self, persistent=self._jac_.persistent)

    def save(self) -> None:
        """Save the node to the memory/storage hierarchy."""
        from jaclang.plugin.feature import JacFeature as Jac

        self._jac_.persistent = True
        Jac.context().save_obj(self, persistent=True)

    def __getstate__(self) -> dict:
        """Override getstate for pickle and shelve."""
        state = self.__dict__.copy()
        state["_jac_"] = self._jac_.__getstate__()
        return state

    def __setstate__(self, state: dict) -> None:
        """Override setstate for pickle and shelve."""
        self.__dict__.update(state)
        self._jac_ = NodeAnchor(obj=self)
        self._jac_.__setstate__(state["_jac_"])


class EdgeArchitype(Architype):
    """Edge Architype Protocol."""

    _jac_: EdgeAnchor
    persistent: bool = False

    def __init__(self) -> None:
        """Create edge architype."""
        from jaclang.plugin.feature import JacFeature as Jac

        self._jac_: EdgeAnchor = EdgeAnchor(obj=self)
        Jac.context().save_obj(self, persistent=self.persistent)

    def save(self) -> None:
        """Save the edge to the memory/storage hierarchy."""
        from jaclang.plugin.feature import JacFeature as Jac

        self.persistent = True
        Jac.context().save_obj(self, persistent=True)

    def __getstate__(self) -> dict:
        """Override getstate for pickle and shelve."""
        state = self.__dict__.copy()
        state["_jac_"] = self._jac_.__getstate__()
        return state

    def __setstate__(self, state: dict) -> None:
        """Override setstate for pickle and shelve."""
        self.__dict__.update(state)
        self._jac_ = EdgeAnchor(obj=self)
        self._jac_.__setstate__(state["_jac_"])

    def populate_nodes(self) -> None:
        """Populate nodes for the edges from node ids."""
        from jaclang.plugin.feature import JacFeature as Jac

        if self._jac_.source_id:
            obj = Jac.context().get_obj(self._jac_.source_id)
            if obj is None:
                raise ValueError(f"Node with id {self._jac_.source_id} not found.")
            elif not isinstance(obj, NodeArchitype):
                raise ValueError(
                    f"Object with id {self._jac_.source_id} is not a node."
                )
            else:
                self._jac_.source = obj
                self._jac_.source_id = None
        if self._jac_.target_id:
            obj = Jac.context().get_obj(self._jac_.target_id)
            if obj is None:
                raise ValueError(f"Node with id {self._jac_.target_id} not found.")
            elif not isinstance(obj, NodeArchitype):
                raise ValueError(
                    f"Object with id {self._jac_.target_id} is not a node."
                )
            else:
                self._jac_.target = obj
                self._jac_.target_id = None


class WalkerArchitype(Architype):
    """Walker Architype Protocol."""

    _jac_: WalkerAnchor

    def __init__(self) -> None:
        """Create walker architype."""
        self._jac_: WalkerAnchor = WalkerAnchor(obj=self)


class GenericEdge(EdgeArchitype):
    """Generic Root Node."""

    _jac_entry_funcs_ = []
    _jac_exit_funcs_ = []


class Root(NodeArchitype):
    """Generic Root Node."""

    _jac_entry_funcs_ = []
    _jac_exit_funcs_ = []
    reachable_nodes: list[NodeArchitype] = []
    connections: set[tuple[NodeArchitype, NodeArchitype, EdgeArchitype]] = set()

    def __init__(self) -> None:
        """Create root node."""
        super().__init__()
        self._jac_.id = UUID(int=0)
        self._jac_.persistent = True

    def reset(self) -> None:
        """Reset the root."""
        self.reachable_nodes = []
        self.connections = set()
        self._jac_.edges = []


@dataclass(eq=False)
class DSFunc:
    """Data Spatial Function."""

    name: str
    trigger: type | types.UnionType | tuple[type | types.UnionType, ...] | None
    func: Callable[[Any, Any], Any] | None = None

    def resolve(self, cls: type) -> None:
        """Resolve the function."""
        self.func = getattr(cls, self.name)



"""Test pass module."""

import jaclang.compiler.absyntree as ast
from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main import DeclImplMatchPass
from jaclang.utils.test import TestCase


class DeclImplMatchPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_ability_connected_to_decl(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("base.jac"), DeclImplMatchPass)
        self.assertFalse(state.errors_had)
        self.assertIn("(o)Test.(c)say_hi", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)say_hi"].decl.name_of.body
        )
        self.assertIn("(o)Test.(c)__init__", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)__init__"].decl.name_of.body
        )

    def test_ability_connected_to_decl_post(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("base2.jac"), DeclImplMatchPass)
        self.assertFalse(state.errors_had)
        self.assertIn("(o)Test.(c)say_hi", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)say_hi"].decl.name_of.body
        )
        self.assertIn("(o)Test.(c)__init__", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)__init__"].decl.name_of.body
        )

    def test_arch_ref_has_sym(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("defs_and_uses.jac"), DeclImplMatchPass
        )
        for i in state.ir.get_all_sub_nodes(ast.ArchRef):
            self.assertIsNotNone(i.sym)



"""Connect Decls and Defs in AST.

This pass creates and manages compilation of Python code from the AST. This pass
also creates bytecode files from the Python code, and manages the caching of
relevant files.
"""

import ast as ast3
import marshal


import jaclang.compiler.absyntree as ast
from jaclang.compiler.passes import Pass


class PyBytecodeGenPass(Pass):
    """Python and bytecode file printing pass."""

    def before_pass(self) -> None:
        """Before pass."""
        return super().before_pass()

    def enter_module(self, node: ast.Module) -> None:
        """Sub objects.

        name: str,
        doc: Token,
        body: Optional['Elements'],
        mod_path: str,
        is_imported: bool,
        sym_tab: Optional[SymbolTable],
        """
        mods = [node] + self.get_all_sub_nodes(node, ast.Module)
        for mod in mods:
            if not mod.gen.py_ast or not isinstance(node.gen.py_ast[0], ast3.Module):
                self.error(
                    f"Unable to find ast for module {node.loc.mod_path}.",
                    node,
                )
                continue
            mod.gen.py_bytecode = marshal.dumps(
                compile(
                    source=mod.gen.py_ast[0], filename=mod.loc.mod_path, mode="exec"
                )
            )
        self.terminate()



"""Main settings of Jac lang."""

import configparser
import os
from dataclasses import dataclass, fields


@dataclass
class Settings:
    """Main settings of Jac lang."""

    # Debug configuration
    fuse_type_info_debug: bool = False
    filter_sym_builtins: bool = True
    ast_symbol_info_detailed: bool = False

    # Import configuration
    py_raise: bool = False
    py_raise_deep: bool = False

    # LSP configuration
    lsp_debug: bool = False

    def __post_init__(self) -> None:
        """Initialize settings."""
        home_dir = os.path.expanduser("~")
        config_dir = os.path.join(home_dir, ".jaclang")
        self.config_file_path = os.path.join(config_dir, "config.ini")
        os.makedirs(config_dir, exist_ok=True)
        if not os.path.exists(self.config_file_path):
            with open(self.config_file_path, "w") as f:
                f.write("[settings]\n")
        self.load_all()

    def load_all(self) -> None:
        """Load settings from all available sources."""
        self.load_config_file()
        self.load_env_vars()

    def load_config_file(self) -> None:
        """Load settings from a configuration file."""
        config_parser = configparser.ConfigParser()
        config_parser.read(self.config_file_path)
        if "settings" in config_parser:
            for key in config_parser["settings"]:
                if key in [f.name for f in fields(self)]:
                    setattr(
                        self, key, self.convert_type(config_parser["settings"][key])
                    )

    def load_env_vars(self) -> None:
        """Override settings from environment variables if available."""
        for key in [f.name for f in fields(self)]:
            env_value = os.getenv("JACLANG_" + key.upper())
            if env_value is not None:
                setattr(self, key, self.convert_type(env_value))

        # def load_command_line_arguments(self):
        #     """Override settings from command-line arguments if provided."""
        #     parser = argparse.ArgumentParser()
        #     parser.add_argument(
        #         "--debug",
        #         type=self.str_to_bool,
        #         nargs="?",
        #         const=True,
        #         default=self.config["debug"],
        #     )
        #     parser.add_argument("--port", type=int, default=self.config["port"])
        #     parser.add_argument("--host", default=self.config["host"])
        #     args = parser.parse_args()

    def str_to_bool(self, value: str) -> bool:
        """Convert string to boolean."""
        return value.lower() in ("yes", "y", "true", "t", "1")

    def convert_type(self, value: str) -> bool | str | int:
        """Convert string values from the config to the appropriate type."""
        if value.isdigit():
            return int(value)
        if value.lower() in (
            "true",
            "false",
            "t",
            "f",
            "yes",
            "no",
            "y",
            "n",
            "1",
            "0",
        ):
            return self.str_to_bool(value)
        return value

    def __str__(self) -> str:
        """Return string representation of the settings."""
        return "\n".join(
            [f"{field.name}: {getattr(self, field.name)}" for field in fields(self)]
        )


settings = Settings()

__all__ = ["settings"]



"""Connect Decls and Defs in AST.

This pass creates links in the ast between Decls of Architypes and Abilities
that are separate from their implementations (Defs). This pass creates a link
in the ast between the Decls and Defs of Architypes and Abilities through the
body field.
"""

import jaclang.compiler.absyntree as ast
from jaclang.compiler.passes import Pass
from jaclang.compiler.passes.main import SubNodeTabPass
from jaclang.compiler.symtable import Symbol, SymbolTable


class DeclImplMatchPass(Pass):
    """Decls and Def matching pass."""

    def enter_module(self, node: ast.Module) -> None:
        """Enter module."""
        if not node.sym_tab:
            self.error(
                f"Expected symbol table on node {node.__class__.__name__}. Perhaps an earlier pass failed."
            )
        else:
            self.connect_def_impl(node.sym_tab)

    def after_pass(self) -> None:
        """Rebuild sub node table."""
        self.ir = SubNodeTabPass(input_ir=self.ir, prior=self).ir

    def defn_lookup(self, lookup: Symbol) -> ast.NameAtom | None:
        """Lookup a definition in a symbol table."""
        for defn in range(len(lookup.defn)):
            candidate = lookup.defn[len(lookup.defn) - (defn + 1)]
            if (
                isinstance(candidate.name_of, ast.AstImplNeedingNode)
                and candidate.name_of.needs_impl
            ):
                return candidate
        return None

    def connect_def_impl(self, sym_tab: SymbolTable) -> None:
        """Connect Decls and Defs."""
        for sym in sym_tab.tab.values():
            if isinstance(sym.decl.name_of, ast.AstImplOnlyNode):
                # currently strips the type info from impls
                arch_refs = [x[3:] for x in sym.sym_name.split(".")]
                name_of_links: list[ast.NameAtom] = []  # to link archref names to decls
                lookup = sym_tab.lookup(arch_refs[0])
                # If below may need to be a while instead of if to skip over local
                # import name collisions (see test: test_impl_decl_resolution_fix)
                if lookup and not isinstance(
                    lookup.decl.name_of, ast.AstImplNeedingNode
                ):
                    lookup = (
                        sym_tab.parent.lookup(arch_refs[0]) if sym_tab.parent else None
                    )
                decl_node = (
                    self.defn_lookup(lookup)
                    if len(arch_refs) == 1 and lookup
                    else lookup.defn[-1] if lookup else None
                )
                name_of_links.append(decl_node) if decl_node else None
                for name in arch_refs[1:]:
                    if decl_node:
                        lookup = (
                            decl_node.name_of.sym_tab.lookup(name, deep=False)
                            if decl_node.name_of.sym_tab
                            else None
                        )
                        decl_node = (
                            self.defn_lookup(lookup)
                            if len(arch_refs) == 1 and lookup
                            else lookup.defn[-1] if lookup else None
                        )
                        name_of_links.append(decl_node) if decl_node else None
                    else:
                        break
                if not decl_node:
                    continue
                elif isinstance(decl_node, ast.Ability) and decl_node.is_abstract:
                    self.warning(
                        f"Abstract ability {decl_node.py_resolve_name()} should not have a definition.",
                        decl_node,
                    )
                    continue
                if not isinstance(
                    valid_decl := decl_node.name_of, ast.AstImplNeedingNode
                ) or not (valid_decl.sym_tab and sym.decl.name_of.sym_tab):
                    raise self.ice(
                        f"Expected AstImplNeedingNode, got {valid_decl.__class__.__name__}. Not possible."
                    )
                valid_decl.body = sym.decl.name_of
                sym.decl.name_of.decl_link = valid_decl
                for idx, a in enumerate(sym.decl.name_of.target.archs):
                    a.name_spec.name_of = name_of_links[idx].name_of
                    a.name_spec.sym = name_of_links[idx].sym
                sym.decl.name_of.sym_tab.tab.update(valid_decl.sym_tab.tab)
                valid_decl.sym_tab.tab = sym.decl.name_of.sym_tab.tab
        for i in sym_tab.kid:
            self.connect_def_impl(i)



"""Transpilation functions."""

from typing import Optional, Type

import jaclang.compiler.absyntree as ast
from jaclang.compiler.parser import JacParser
from jaclang.compiler.passes import Pass
from jaclang.compiler.passes.main import PyOutPass, pass_schedule
from jaclang.compiler.passes.tool import JacFormatPass
from jaclang.compiler.passes.tool.schedules import format_pass


def compile_jac(file_path: str, cache_result: bool = False) -> Pass:
    """Start Compile for Jac file and return python code as string."""
    code = jac_file_to_pass(
        file_path=file_path,
        schedule=pass_schedule,
    )
    if cache_result and isinstance(code.ir, ast.Module):
        print_pass = PyOutPass(input_ir=code.ir, prior=code)
        return print_pass
    else:
        return code


def jac_file_to_pass(
    file_path: str,
    target: Optional[Type[Pass]] = None,
    schedule: list[Type[Pass]] = pass_schedule,
) -> Pass:
    """Convert a Jac file to an AST."""
    with open(file_path) as file:
        return jac_str_to_pass(
            jac_str=file.read(),
            file_path=file_path,
            target=target,
            schedule=schedule,
        )


def jac_str_to_pass(
    jac_str: str,
    file_path: str,
    target: Optional[Type[Pass]] = None,
    schedule: list[Type[Pass]] = pass_schedule,
) -> Pass:
    """Convert a Jac file to an AST."""
    if not target:
        target = schedule[-1] if schedule else None
    source = ast.JacSource(jac_str, mod_path=file_path)
    ast_ret: Pass = JacParser(input_ir=source)
    for i in schedule:
        if i == target:
            break
        ast_ret = i(input_ir=ast_ret.ir, prior=ast_ret)
    ast_ret = target(input_ir=ast_ret.ir, prior=ast_ret) if target else ast_ret
    return ast_ret


def jac_ir_to_pass(
    ir: ast.AstNode,
    target: Optional[Type[Pass]] = None,
    schedule: list[Type[Pass]] = pass_schedule,
) -> Pass:
    """Convert a Jac file to an AST."""
    if not target:
        target = schedule[-1] if schedule else None
    ast_ret = (
        Pass(input_ir=ir, prior=None)
        if not len(schedule)
        else schedule[0](input_ir=ir, prior=None)
    )
    for i in schedule[1:]:
        if i == target:
            break
        ast_ret = i(input_ir=ast_ret.ir, prior=ast_ret)
    ast_ret = target(input_ir=ast_ret.ir, prior=ast_ret) if target else ast_ret
    return ast_ret


def jac_pass_to_pass(
    in_pass: Pass,
    target: Optional[Type[Pass]] = None,
    schedule: list[Type[Pass]] = pass_schedule,
) -> Pass:
    """Convert a Jac file to an AST."""
    if not target:
        target = schedule[-1] if schedule else None
    ast_ret = in_pass
    for i in schedule:
        if i == target:
            break
        ast_ret = i(input_ir=ast_ret.ir, prior=ast_ret)
    ast_ret = target(input_ir=ast_ret.ir, prior=ast_ret) if target else ast_ret
    return ast_ret


def jac_file_formatter(
    file_path: str,
    schedule: list[Type[Pass]] = format_pass,
) -> JacFormatPass:
    """Convert a Jac file to an AST."""
    target = JacFormatPass
    with open(file_path) as file:
        source = ast.JacSource(file.read(), mod_path=file_path)
        prse: Pass = JacParser(input_ir=source)
    for i in schedule:
        if i == target:
            break
        prse = target(input_ir=prse.ir, prior=prse)
    prse = target(input_ir=prse.ir, prior=prse)
    return prse



"""Jac to python ast link pass.

This pass is needed so cases where there are multiple Jac nodes relevant to a
single python node can be linked. For example FuncDef doesn't have a Name node
however Ability does.
"""

import ast as ast3

import jaclang.compiler.absyntree as ast
from jaclang.compiler.passes import Pass


class PyJacAstLinkPass(Pass):
    """Link jac ast to python ast nodes."""

    def link_jac_py_nodes(
        self, jac_node: ast.AstNode, py_nodes: list[ast3.AST]
    ) -> None:
        """Link jac name ast to py ast nodes."""
        jac_node.gen.py_ast = py_nodes
        for i in py_nodes:
            if isinstance(i.jac_link, list):  # type: ignore
                i.jac_link.append(jac_node)  # type: ignore

    def exit_module_path(self, node: ast.ModulePath) -> None:
        """Sub objects.

        path: Sequence[Token],
        alias: Optional[Name],
        path_str: str,
        """
        if node.alias:
            self.link_jac_py_nodes(jac_node=node.alias, py_nodes=node.gen.py_ast)

    def exit_architype(self, node: ast.Architype) -> None:
        """Sub objects.

        name: Name,
        arch_type: Token,
        access: Optional[SubTag[Token]],
        base_classes: Optional[SubNodeList[AtomType]],
        body: Optional[SubNodeList[ArchBlockStmt] | ArchDef],
        doc: Optional[String],
        decorators: Optional[SubNodeList[ExprType]],
        """
        self.link_jac_py_nodes(jac_node=node.name, py_nodes=node.gen.py_ast)
        if isinstance(node.body, ast.ArchDef):
            self.link_jac_py_nodes(jac_node=node.body, py_nodes=node.gen.py_ast)

    def exit_arch_def(self, node: ast.ArchDef) -> None:
        """Sub objects.

        target: ArchRefChain,
        body: SubNodeList[ArchBlockStmt],
        doc: Optional[String],
        decorators: Optional[SubNodeList[ExprType]],
        """
        for i in node.target.archs:
            if i.name_spec.name_of.sym:
                self.link_jac_py_nodes(
                    jac_node=i, py_nodes=i.name_spec.name_of.sym.decl.gen.py_ast
                )
                self.link_jac_py_nodes(
                    jac_node=i.name_spec,
                    py_nodes=i.name_spec.name_of.sym.decl.gen.py_ast,
                )

    def exit_enum(self, node: ast.Enum) -> None:
        """Sub objects.

        name: Name,
        access: Optional[SubTag[Token]],
        base_classes: Optional[Optional[SubNodeList[AtomType]]],
        body: Optional[SubNodeList[EnumBlockStmt] | EnumDef],
        doc: Optional[String],
        decorators: Optional[SubNodeList[ExprType]],
        """
        if isinstance(node.body, ast.EnumDef):
            self.link_jac_py_nodes(jac_node=node.body, py_nodes=node.gen.py_ast)

    def exit_enum_def(self, node: ast.EnumDef) -> None:
        """Sub objects.

        target: ArchRefChain,
        body: SubNodeList[EnumBlockStmt],
        doc: Optional[String],
        decorators: Optional[SubNodeList[ExprType]],
        """
        for i in node.target.archs:
            if i.name_spec.name_of.sym:
                self.link_jac_py_nodes(
                    jac_node=i, py_nodes=i.name_spec.name_of.sym.decl.gen.py_ast
                )
                self.link_jac_py_nodes(
                    jac_node=i.name_spec,
                    py_nodes=i.name_spec.name_of.sym.decl.gen.py_ast,
                )

    def exit_ability(self, node: ast.Ability) -> None:
        """Sub objects.

        name_ref: NameType,
        is_func: bool,
        is_async: bool,
        is_static: bool,
        is_abstract: bool,
        access: Optional[SubTag[Token]],
        signature: Optional[FuncSignature | ExprType | EventSignature],
        body: Optional[SubNodeList[CodeBlockStmt] | AbilityDef | FuncCall],
        doc: Optional[String],
        decorators: Optional[SubNodeList[ExprType]],
        """
        self.link_jac_py_nodes(jac_node=node.name_ref, py_nodes=node.gen.py_ast)
        if isinstance(node.body, ast.AbilityDef):
            self.link_jac_py_nodes(jac_node=node.body, py_nodes=node.gen.py_ast)

    def exit_ability_def(self, node: ast.AbilityDef) -> None:
        """Sub objects.

        target: ArchRefChain,
        signature: FuncSignature | EventSignature,
        body: SubNodeList[CodeBlockStmt],
        doc: Optional[String],
        decorators: Optional[SubNodeList[ExprType]],
        """
        for i in node.target.archs:
            if i.name_spec.name_of.sym:
                self.link_jac_py_nodes(
                    jac_node=i, py_nodes=i.name_spec.name_of.sym.decl.gen.py_ast
                )
                self.link_jac_py_nodes(
                    jac_node=i.name_spec,
                    py_nodes=i.name_spec.name_of.sym.decl.gen.py_ast,
                )

        if isinstance(node.parent, ast.Ability) and node.parent.signature:
            if isinstance(node.signature, ast.FuncSignature) and node.signature.params:
                for src_prm in node.signature.params.items:
                    if (
                        isinstance(node.parent.signature, ast.FuncSignature)
                        and node.parent.signature.params
                    ):
                        for trg_prm in node.parent.signature.params.items:
                            if src_prm.name.sym_name == trg_prm.name.sym_name:
                                self.link_jac_py_nodes(
                                    jac_node=src_prm, py_nodes=trg_prm.gen.py_ast
                                )
                                self.link_jac_py_nodes(
                                    jac_node=src_prm.name,
                                    py_nodes=trg_prm.name.gen.py_ast,
                                )
            if (
                isinstance(node.signature, ast.FuncSignature)
                and node.signature.return_type
            ) and (
                isinstance(node.parent.signature, ast.FuncSignature)
                and node.parent.signature.return_type
            ):
                self.link_jac_py_nodes(
                    jac_node=node.signature.return_type,
                    py_nodes=node.parent.signature.return_type.gen.py_ast,
                )

    def exit_param_var(self, node: ast.ParamVar) -> None:
        """Sub objects.

        name: Name,
        unpack: Optional[Token],
        type_tag: SubTag[ExprType],
        value: Optional[ExprType],
        """
        self.link_jac_py_nodes(jac_node=node.name, py_nodes=node.gen.py_ast)

    def exit_except(self, node: ast.Except) -> None:
        """Sub objects.

        ex_type: ExprType,
        name: Optional[Name],
        body: SubNodeList[CodeBlockStmt],
        """
        if node.name:
            self.link_jac_py_nodes(jac_node=node.name, py_nodes=node.gen.py_ast)

    def exit_expr_as_item(self, node: ast.ExprAsItem) -> None:
        """Sub objects.

        expr: ExprType,
        alias: Optional[ExprType],
        """
        self.link_jac_py_nodes(jac_node=node.expr, py_nodes=node.gen.py_ast)
        if node.alias:
            self.link_jac_py_nodes(jac_node=node.alias, py_nodes=node.gen.py_ast)

    def exit_global_stmt(self, node: ast.GlobalStmt) -> None:
        """Sub objects.

        target: SubNodeList[NameType],
        """
        for x, y in enumerate(node.target.items):
            self.link_jac_py_nodes(jac_node=y, py_nodes=[node.gen.py_ast[x]])

    def exit_non_local_stmt(self, node: ast.NonLocalStmt) -> None:
        """Sub objects.

        target: SubNodeList[NameType],
        """
        for x, y in enumerate(node.target.items):
            self.link_jac_py_nodes(jac_node=y, py_nodes=[node.gen.py_ast[x]])

    def exit_k_w_pair(self, node: ast.KWPair) -> None:
        """Sub objects.

        key: NameType,
        value: ExprType,
        """
        if node.key:
            self.link_jac_py_nodes(jac_node=node.key, py_nodes=node.gen.py_ast)

    def exit_atom_trailer(self, node: ast.AtomTrailer) -> None:
        """Sub objects.

        target: Expr,
        right: AtomExpr | Expr,
        is_attr: bool,
        is_null_ok: bool,
        is_genai: bool = False,
        """
        if node.is_attr and isinstance(node.right, ast.AstSymbolNode):
            self.link_jac_py_nodes(
                jac_node=node.right.name_spec, py_nodes=node.gen.py_ast
            )



"""Pass schedules.

These are various pass schedules for the Jac compiler and static analysis.
"""

from __future__ import annotations


from .sub_node_tab_pass import SubNodeTabPass  # noqa: I100
from .import_pass import JacImportPass, PyImportPass  # noqa: I100
from .sym_tab_build_pass import SymTabBuildPass  # noqa: I100
from .def_impl_match_pass import DeclImplMatchPass  # noqa: I100
from .def_use_pass import DefUsePass  # noqa: I100
from .pyout_pass import PyOutPass  # noqa: I100
from .pybc_gen_pass import PyBytecodeGenPass  # noqa: I100
from .pyast_gen_pass import PyastGenPass  # noqa: I100
from .pyjac_ast_link_pass import PyJacAstLinkPass  # noqa: I100
from .type_check_pass import JacTypeCheckPass  # noqa: I100
from .fuse_typeinfo_pass import FuseTypeInfoPass  # noqa: I100
from .registry_pass import RegistryPass  # noqa: I100
from .access_modifier_pass import AccessCheckPass  # noqa: I100

py_code_gen = [
    SubNodeTabPass,
    JacImportPass,
    PyImportPass,
    SymTabBuildPass,
    DeclImplMatchPass,
    DefUsePass,
    RegistryPass,
    PyastGenPass,
    PyJacAstLinkPass,
    PyBytecodeGenPass,
]

type_checker_sched = [JacTypeCheckPass, FuseTypeInfoPass, AccessCheckPass]
py_code_gen_typed = [*py_code_gen, *type_checker_sched]
py_compiler = [*py_code_gen, PyOutPass]



# flake8: noqa

"""Python function for testing py imports."""

global a, X, y, z
a = 5
X = 10
y = 15
z = 20

print(a, X, y, z)

x = 15
if 0 <= x <= 5:
    print("Not Bad")
elif 6 <= x <= 10:
    print("Average")
else:
    print("Good Enough")

from enum import Enum


def foo() -> None:
    return "Hello"


class vehicle:
    def __init__(self) -> None:
        self.name = "Car"


class Size(Enum):
    Small = 1
    Medium = 2
    Large = 3


car = vehicle()
print(foo())
print(car.name)
print(Size.Medium.value)


from jaclang import jac_import as __jac_import__
import os
from math import sqrt as square_root
import datetime as dt

for i in range(int(square_root(dt.datetime.now().year))):
    print(os.getcwd())

print("hello ")


def foo():
    print("world")


foo()

x = lambda a, b: a + b
print(x(5, 4))

if 5 > 4:
    print("True")
elif "a" != "a":
    print("'a' is 'a' ")
else:
    print("False")
a = [1, 2, 3]
b = [1, 2, 3]
print(a is b)
print(3 in a)

print(True or False)
print(False and False)

day = "sunday"
match day:
    case "monday":
        print("confirmed")
    case _ as day:
        print("other")

num = 89
match num:
    case 89:
        print("Correct")
    case 8:
        print("Nope")

data = {"key1": 1, "key2": 2, "232": 3453}

match data:
    case {"key1": 1, "key2": 2, **rest}:
        print(f"Matched a mapping with key1 and key2. Rest: {rest}")


class Point:
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y


def match_example(data: any):
    match data:
        # MatchValue
        case 42:
            print("Matched the value 42.")

        # MatchSingleton
        case True:
            print("Matched the singleton True.")
        case None:
            print("Matched the singleton None.")

        # MatchSequence
        case [1, 2, 3]:
            print("Matched a specific sequence [1, 2, 3].")

        # MatchStar
        case [1, *rest, 3]:
            print(
                f"Matched a sequence starting with 1 and ending with 3. Middle: {rest}"
            )

        # MatchMapping
        case {"key1": 1, "key2": 2, **rest}:
            print(f"Matched a mapping with key1 and key2. Rest: {rest}")

        # MatchClass
        case Point(x=int(a), y=0):
            print(f"Point with x={a} and y=0")

        # MatchAs
        case [1, 2, rest_val as value]:
            print(f"Matched a sequence and captured the last value: {value}")

        # MatchOr
        case [1, 2] | [3, 4]:
            print("Matched either the sequence [1, 2] or [3, 4].")

        # case _:
        #     print("No match found.")


match_example(Point(x=9, y=0))

data = [1, 2, 3]
match data:
    case [1, 2, 3]:
        print("Matched")
    case _:
        print("Not Found")
data = True
match True:
    # MatchSingleton
    case True:
        print("Matched the singleton True.")
    case None:
        print("Matched the singleton None.")
a = 8
match a:
    case 7:
        print("Doable")
    case _:
        print("Undoable")


class Animal:
    def __init__(self, animal_type):
        self.animal_type = animal_type


class Dog(Animal):
    def __init__(self, breed):
        super().__init__("Dog")
        self.breed = breed
        self.sound = ""
        self._post_init()

    def _post_init(self):
        self.sound = "woof woof"


dog1 = Dog(breed="Labrador")
print(dog1.animal_type, dog1.breed, dog1.sound)


def double(x: int) -> int:
    return x * 2


number = 5
result = double(number)
print(result)


def square(x: int) -> int:
    return x**2


number = 5
result = square(number)
print(result)


def foo(value: int):
    if value < 0:
        raise ValueError("Value must be non-negative")


try:
    foo(-1)
except ValueError as e:
    print("Raised:", e)


try:
    with open("test.txt") as file:
        print(file.read())
except:
    print("File not found")

try:
    with open("test.txt") as file:
        print(file.read())
except FileNotFoundError:
    print("File not found")

if 5 < 0:
    raise ValueError("A value error occurred")
elif 3 < 1:
    try:
        x = 1 / 0
    except ZeroDivisionError as e:
        raise ValueError("Error occurred") from e
else:
    print("No error occurred")


def foo():
    a = 42
    if a > 0:
        return
    return a


print("Returned:", foo())


class Sample:
    def __init__(self):
        self.my_list = [1, 2, 3]
        self.my_dict = {"name": "John", "age": 30}


def main():
    sample_instance = Sample()
    first, second = sample_instance.my_list[2], sample_instance.my_dict["name"]

    print(first, second)


main()


import unittest


class TestCases(unittest.TestCase):
    def test_test1(self):
        self.assertAlmostEqual(4.99999, 4.99999)

    def test_test2(self):
        self.assertEqual(5, 5)

    def test_test3(self):
        self.assertIn("e", "qwerty")


if __name__ == "__main__":
    unittest.main()


def foo(first: int, second: int) -> None:
    print(first, second)


val1 = (3,) + (4,)
val2 = (val1[0] * val1[1], val1[0] + val1[1])
foo(second=val2[1], first=val2[0])
foo(first=val2[0], second=val2[1])


def combine_via_func(a: int, b: int, c: int, d: int) -> int:
    return a + b + c + d


def myFunc():
    yield "Hello"
    yield 91
    yield "Good Bye"


x = myFunc()

for z in x:
    print(z)



"""Tests for Jac parser."""

import inspect

from jaclang.compiler import jac_lark as jl
from jaclang.compiler.absyntree import JacSource
from jaclang.compiler.constant import Tokens
from jaclang.compiler.parser import JacParser
from jaclang.utils.test import TestCaseMicroSuite


class TestLarkParser(TestCaseMicroSuite):
    """Test Jac self.prse."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_fstring_escape_brace(self) -> None:
        """Test fstring escape brace."""
        source = JacSource('glob a=f"{{}}", not_b=4;', mod_path="")
        prse = JacParser(input_ir=source)
        self.assertFalse(prse.errors_had)

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        prse = JacParser(
            input_ir=JacSource(self.file_to_str(filename), mod_path=filename),
        )
        self.assertFalse(prse.errors_had)

    def test_parser_fam(self) -> None:
        """Parse micro jac file."""
        prse = JacParser(input_ir=JacSource(self.load_fixture("fam.jac"), mod_path=""))
        self.assertFalse(prse.errors_had)

    def test_staticmethod_checks_out(self) -> None:
        """Parse micro jac file."""
        prse = JacParser(
            input_ir=JacSource(
                self.load_fixture("staticcheck.jac"),
                mod_path="",
            )
        )
        out = prse.ir.pp()
        self.assertFalse(prse.errors_had)
        self.assertNotIn("staticmethod", out)

    def test_parser_kwesc(self) -> None:
        """Parse micro jac file."""
        prse = JacParser(
            input_ir=JacSource(self.load_fixture("kwesc.jac"), mod_path="")
        )
        self.assertFalse(prse.errors_had)

    def test_parser_mod_doc_test(self) -> None:
        """Parse micro jac file."""
        prse = JacParser(
            input_ir=JacSource(self.load_fixture("mod_doc_test.jac"), mod_path="")
        )
        self.assertFalse(prse.errors_had)

    def test_enum_matches_lark_toks(self) -> None:
        """Test that enum stays synced with lexer."""
        tokens = [x.name for x in jl.Lark_StandAlone().parser.lexer_conf.terminals]
        for token in tokens:
            self.assertIn(token, Tokens.__members__)
        for token in Tokens:
            self.assertIn(token.name, tokens)
        for token in Tokens:
            self.assertIn(token.value, tokens)

    def test_parser_impl_all_rules(self) -> None:
        """Test that enum stays synced with lexer."""
        rules = {
            x.origin.name
            for x in jl.Lark_StandAlone().parser.parser_conf.rules
            if not x.origin.name.startswith("_")
        }
        parse_funcs = []
        for name, value in inspect.getmembers(JacParser.TreeToAST):
            if inspect.isfunction(value) and not getattr(
                JacParser.TreeToAST.__base__, value.__name__, False
            ):
                parse_funcs.append(name)
        for i in rules:
            self.assertIn(i, parse_funcs)
        for i in parse_funcs:
            if i in ["binary_expr_unwind", "ice", "nu"]:
                continue
            self.assertIn(i, rules)

    def test_all_ast_has_normalize(self) -> None:
        """Test for enter/exit name diffs with parser."""
        import jaclang.compiler.absyntree as ast
        import inspect
        import sys

        exclude = [
            "AstNode",
            "WalkerStmtOnlyNode",
            "JacSource",
            "EmptyToken",
            "AstSymbolNode",
            "AstSymbolStubNode",
            "AstImplNeedingNode",
            "AstAccessNode",
            "Literal",
            "AstDocNode",
            "AstSemStrNode",
            "PythonModuleAst",
            "AstAsyncNode",
            "AstElseBodyNode",
            "AstTypedVarNode",
            "AstImplOnlyNode",
            "Expr",
            "AtomExpr",
            "ElementStmt",
            "ArchBlockStmt",
            "EnumBlockStmt",
            "CodeBlockStmt",
            "NameAtom",
            "ArchSpec",
            "MatchPattern",
        ]
        module_name = ast.__name__
        module = sys.modules[module_name]

        # Retrieve the source code of the module
        source_code = inspect.getsource(module)

        classes = inspect.getmembers(module, inspect.isclass)
        ast_node_classes = [
            cls
            for _, cls in classes
            if issubclass(cls, ast.AstNode) and not issubclass(cls, ast.Token)
        ]

        ordered_classes = sorted(
            ast_node_classes, key=lambda cls: source_code.find(f"class {cls.__name__}")
        )
        for cls in ordered_classes:
            if cls.__name__ not in exclude:
                self.assertIn("normalize", cls.__dict__)


TestLarkParser.self_attach_micro_tests()



"""Collection of passes for Jac IR."""

from .sub_node_tab_pass import SubNodeTabPass
from .import_pass import JacImportPass, PyImportPass  # noqa: I100
from .sym_tab_build_pass import SymTabBuildPass  # noqa: I100
from .def_impl_match_pass import DeclImplMatchPass  # noqa: I100
from .def_use_pass import DefUsePass  # noqa: I100
from .pyout_pass import PyOutPass  # noqa: I100
from .pyast_load_pass import PyastBuildPass  # type: ignore # noqa: I100
from .pyast_gen_pass import PyastGenPass  # noqa: I100
from .schedules import py_code_gen  # noqa: I100
from .type_check_pass import JacTypeCheckPass  # noqa: I100
from .registry_pass import RegistryPass  # noqa: I100


pass_schedule = py_code_gen

__all__ = [
    "SubNodeTabPass",
    "JacImportPass",
    "PyImportPass",
    "SymTabBuildPass",
    "DeclImplMatchPass",
    "DefUsePass",
    "PyOutPass",
    "PyastBuildPass",
    "PyastGenPass",
    "JacTypeCheckPass",
    "RegistryPass",
]



"""Test pass module."""

# from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main import SymTabBuildPass
from jaclang.utils.test import AstSyncTestMixin, TestCase


class SymTabBuildPassTests(TestCase, AstSyncTestMixin):
    """Test pass module."""

    TargetPass = SymTabBuildPass

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    # def test_name_collision(self) -> None:
    #     """Basic test for pass."""
    #     state = jac_file_to_pass(
    #         self.fixture_abs_path("multi_def_err.jac"), SymTabBuildPass
    #     )
    #     self.assertGreater(len(state.warnings_had), 0)
    #     self.assertIn("MyObject", str(state.warnings_had[0]))
